#! /usr/bin/env python
#
#   gregress.py -- Front-end script for running REMDataset pipeline.
#
#   Copyright (C) 2017, 2018  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2017-07-03:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '1.0.0'

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gregress
    gregress.GenREMScript().run()

import os
import sys
import shutil
import random
from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, \
    existing_directory
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, \
    hours, minutes, seconds
from gc3libs.workflow import StagedTaskCollection, \
    ParallelTaskCollection, SequentialTaskCollection

# Defaults

S0_REMOTE_INPUT_FILENAME = "./input.stata"
S0_REMOTE_OUTPUT_FILENAME = "./output.RData"
RSCRIPT_COMMAND="Rscript --vanilla {src}/script.R {method} {src} {data}"
STATS = ["stat-inertia", "stat-reciprocity", "stat-similarity", "stat-triad", "stat-degree"]
METHODS = ['remdataset', 'merge'] + STATS
S0_OUTPUT=""
S1_OUTPUT=""
S2_OUTPUT=""
REMOTE_RESULT_FOLDER="result"
REMOTE_DATA_FOLDER="data"
REMOTE_SCRIPTS_FOLDER="src"




# Utility methods

class GenREMDatasetApplication(Application):
    """
    Execute main workflow script and pass the workflow step
    as input argument.
    """
    application_name = 'genREM'

    def __init__(self, method, data_file_list, source_folder, **extra_args):

        inputs = dict()

        self.output = extra_args['results']

        for data_folder in data_file_list:
            for data_file in os.listdir(data_folder):
                inputs[os.path.join(data_folder,
                                    data_file)] = os.path.join(REMOTE_DATA_FOLDER,
                                                               data_file)

        for script_file in os.listdir(source_folder):
            inputs[os.path.join(source_folder,
                                script_file)] = os.path.join(REMOTE_SCRIPTS_FOLDER,
                                                             os.path.basename(script_file))

        arguments = RSCRIPT_COMMAND.format(method=method,
                                           src=REMOTE_SCRIPTS_FOLDER,
                                           data=REMOTE_DATA_FOLDER)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = [REMOTE_RESULT_FOLDER],
            stdout = 'genREM.log',
            join=True,
            **extra_args)

    def terminated(self):
        """
        Move results to corresponding 'result' folder
        """

        if not os.path.isdir(self.output):
            os.makedirs(self.output)
        for data in os.listdir(os.path.join(self.output_dir,
                                            REMOTE_RESULT_FOLDER)):
            shutil.move(os.path.join(self.output_dir,
                                     REMOTE_RESULT_FOLDER,
                                     data),
                        os.path.join(self.output,
                                     data))

class GenREMStagedTaskCollection(StagedTaskCollection):
    """
    Staged collection:
    Step 0: Generate REMDataset
    Step 1: For each available statistical method, run independent application
    Step 2: Merge all results together
    """
    def __init__(self, data_folder, source_folder, **extra_args):

        self.data_folder = data_folder
        self.source_folder = source_folder
        self.extra = extra_args
        self.s0_outputfolder = os.path.join(extra_args['result'],"S0")
        self.s1_outputfolder = os.path.join(extra_args['result'],"S1")
        self.s2_outputfolder = os.path.join(extra_args['result'],"S2")
        StagedTaskCollection.__init__(self)

    def stage0(self):
        """
        Step 0: Generate REMDataset
        """

        extra_args = self.extra.copy()
        extra_args['jobname'] = "remdataset"
        extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                    extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                    extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                    extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                    extra_args['jobname'])

        # gc3libs.log.debug("Creating Stage0 task for : %s" % os.path.basename(self.input_stata_file))
        extra_args['results'] = self.s0_outputfolder

        return GenREMDatasetApplication("remdataset",[self.data_folder],self.source_folder,**extra_args)


    def stage1(self):
        """
        Step 1: For each available statistical method, run independent application
        """
        tasks = []

        for method in STATS:
            extra_args = self.extra.copy()
            extra_args['jobname'] = method
            extra_args['results'] = self.s1_outputfolder
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        extra_args['jobname'])

            tasks.append(GenREMDatasetApplication(method,[self.s0_outputfolder],self.source_folder,**extra_args))
        return ParallelTaskCollection(tasks)

    def stage2(self):
        """
        Step 2: Merge all results together
        """
        extra_args = self.extra.copy()
        extra_args['jobname'] = "merge"
        extra_args['results'] = self.s2_outputfolder
        extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                    extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                    extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                    extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                    extra_args['jobname'])

        return GenREMDatasetApplication("merge",[self.s0_outputfolder, self.s1_outputfolder],self.source_folder,**extra_args)


class GenREMScript(SessionBasedScript):
    """
    Take initial RData DataFrame input file and run the full workflow by
    passing to the main execution script the workflow steps as part of the input
    arguments.

    The ``gregress`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gscr`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GenREMDatasetApplication,
            stats_only_for = GenREMDatasetApplication
            )

    def setup_args(self):

        self.add_param('data_folder', type=existing_directory,
                       help="Location of initial RData file.")

    def setup_options(self):
        self.add_param("-x", "--source", metavar="PATH",
                       dest="src", default=None,
                       type=existing_directory,
                       help="Location of source R scripts.")

        self.add_param("-R", "--results", metavar="PATH",
                       dest="result", default='results',
                       help="Location of results.")

    def new_tasks(self, extra):
        """
        Just start the StageCollection with the passed input RData file in
        self.params.data_folder.
        """
        extra_args = extra.copy()
        extra_args['result'] = self.params.result
        return [GenREMStagedTaskCollection(self.params.data_folder,
                                           self.params.src,
                                           **extra_args)]



#! /usr/bin/env python
#
#   gsceuafish.py -- Front-end script for evaluating Matlab functions
#   function over a large number of parameters.
#
#   Copyright (C) 2015, 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple Matlab jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gsceuafish.py --help`` for program usage
instructions.

Input parameters consists of:
:param str input_csv: Path to an .csv file containing input parameters
Example:
1,2,1,4
1,1,1,1
...
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-11-05:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gsceuafish
    gsceuafish.GsceuafishScript().run()

import os
import sys
import time
import tempfile

import shutil
import pandas

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

## custom application class
class GsceuafishApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'sceuafish'

    def __init__(self, parameter, **extra_args):

        inputs = dict()
        outputs = dict()

        # execution wrapper needs to be added anyway
        gsceuafish_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gsceuafish.sh")
        inputs[gsceuafish_wrapper_sh] = os.path.basename(gsceuafish_wrapper_sh)

        _command = "./%s %s " % (os.path.basename(gsceuafish_wrapper_sh),
                                 ' '.join(str(x) for x in parameter))

        if "main_loop_folder" in extra_args:
            inputs[extra_args['main_loop_folder']] = './data/'

        Application.__init__(
            self,
            arguments = _command,
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'gsceuafish.log',
            join=True,
            executables = "./%s" % os.path.basename(gsceuafish_wrapper_sh),
            **extra_args)


class GsceuafishScript(SessionBasedScript):
    """
    Parse the input .csv file. For each line in the input .csv file
    create a different Matlab execution.

    The ``gsceuafish`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gsceuafish`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gsceuafish``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GsceuafishApplication,
            stats_only_for = GsceuafishApplication,
            )

    def setup_options(self):
        self.add_param("-d", "--data", metavar="PATH", type=str,
                       dest="main_loop", default=None,
                       help="Location of the Main_Loop.m script and "
                       "related MAtlab functions. Default: None")

    def setup_args(self):

        self.add_param('csv_input_file', type=str,
                       help="Input .csv file")

    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """
        assert os.path.isfile(self.params.csv_input_file), \
        "Input CSV file %s not found" % self.params.csv_input_file

        if self.params.main_loop:
            assert os.path.isdir(self.params.main_loop), \
            "Main_Loop.m location %s not found" % self.params.main_loop

    def new_tasks(self, extra):
        """
        For each line of the input .csv file generate
        an execution Task
        """
        tasks = []

        for parameter in self._enumerate_csv(self.params.csv_input_file):
            parameter_str = '.'.join(str(x) for x in parameter)
            jobname = "gsceuafish-%s" % parameter_str

            extra_args = extra.copy()

            extra_args['jobname'] = jobname

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

            if self.params.main_loop:
                extra_args['main_loop_folder'] = self.params.main_loop

            self.log.debug("Creating Application for parameter : %s" %
                           (parameter_str))

            tasks.append(GsceuafishApplication(
                    parameter,
                    **extra_args))

        return tasks

    def _enumerate_csv(self, input_csv):
        """
        For each line of the input .csv file
        return list of parameters
        """
        parameters = pandas.read_csv(input_csv,header=None)
        for i,p in enumerate(parameters.values):
            yield p.tolist()



#! /usr/bin/env python
#
"""
Housing-specific methods and overloads.
"""
# Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

__author__ = 'Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
# summary of user-visible changes
__changelog__ = """

"""
__docformat__ = 'reStructuredText'

from __future__ import absolute_import, print_function
import gc3libs.debug
import gc3libs.application.apppot
import re, os
import numpy as np
from supportGc3 import lower, flatten, str2tuple, getIndex, extractVal, str2vals
from supportGc3 import format_newVal, update_parameter_in_file, safe_eval, str2mat, mat2str, getParameter
from paraLoop import paraLoop

from gc3libs import Application, Run
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
import shutil

import logbook, sys
from supportGc3 import wrapLogger
# import personal libraries
path2SrcPy = os.path.join(os.path.dirname(__file__), '../src')
if not sys.path.count(path2SrcPy):
    sys.path.append(path2SrcPy)
from plotSimulation import plotSimulation
from plotAggregate import makeAggregatePlot
from pymods.classes.tableDict import tableDict
from plotOwnership import plotOwnerProfiles

path2Pymods = os.path.join(os.path.dirname(__file__), '../')
if not sys.path.count(path2Pymods):
    sys.path.append(path2Pymods)
from pymods.support.support import getParameter

logger = wrapLogger(loggerName = __name__ + 'logger', streamVerb = 'DEBUG', logFile = __name__ + '.log')

class housingApplication(Application):

    application_name = 'housing'

    _invalid_chars = re.compile(r'[^_a-zA-Z0-9]+', re.X)

    def __init__(self, executable, arguments, inputs, outputs, output_dir, **extra_args):
#        extra_args.setdefault('requested_walltime', 2*hours)
        Application.__init__(self, executable, arguments, inputs, outputs, output_dir, **extra_args)

    def fetch_output_error(self, ex):

        if self.execution.state == Run.State.TERMINATING:
        # do notify task/main application that we're done
        # ignore error, let's continue
            self.execution.state = Run.State.TERMINATED
            logger.debug('fetch_output_error occured... continuing')
            if self.persistent_id:
                logger.debug('jobid: %s exception: %s' % (self.persistent_id, str(ex)))
            else:
                logger.debug('info: %s exception: %s' % (self.info, str(ex)))
            return None
        else:
        # non-terminal state, pass on error
            return ex

    # def submit_error(self, ex):
    #     logger.debug('submit_error occured... continuing')

    #     try:
    #         if self.lrms_jobid:
    #             logger.debug('jobid: %s info: %s exception: %s' % (self.lrms_jobid, self.info, str(ex)))
    #         else:
    #             logger.debug('info: %s exception: %s' % (self.info, str(ex)))
    #     except AttributeError:
    #         logger.debug('no `lrms_jobid` hence submission didnt happen')
    #     return None


    def terminated(self):
        """
        Analyze the retrieved output, with a threefold purpose:

        - work around a bug in ARC where the output is stored in a
          subdirectory of the output directory.

        - make plots for post-analysis
        """
        output_dir = self.output_dir
        # if files are stored in `output/output/`, move them one level up
        if os.path.isdir(os.path.join(output_dir, 'output')):
            wrong_output_dir = os.path.join(output_dir, 'output')
            for entry in os.listdir(wrong_output_dir):
                dest_entry = os.path.join(output_dir, entry)
                if os.path.isdir(dest_entry):
                    # backup with numerical suffix
                    gc3libs.utils.backup(dest_entry)
                os.rename(os.path.join(wrong_output_dir, entry), dest_entry)
            try:
                os.rmdir(wrong_output_dir)
            except:
                logger.warning('could not delete wront output dir = %s' % wrong_output_dir)


        ## set the exitcode based on postprocessing the main output file
        #aggregateOut = os.path.join(output_dir, 'aggregate.out')
        #genParametersFile = os.path.join(os.getcwd(), 'localBaseDir', 'input', 'genParameters.in')
        #ctry =       getParameter(genParametersFile, 'ctry', 'space-separated')
        #if ctry == 'us':
            #curPanel = 'PSID'
        #elif ctry == 'de':
            #curPanel = 'SOEP'
        #elif ctry == 'uk':
            #curPanel = 'BHPS'
        #else:
            #logger.critical('unknown profile %s' % profile)
            #os.exit(1)
        #empOwnershipFile = os.path.join(os.path.split(output_dir)[0], 'input', curPanel + 'OwnershipProfilealleduc.out')
        #ownershipTableFile = os.path.join(output_dir, 'ownershipTable.out')
        #if os.path.exists(aggregateOut):
            #self.execution.exitcode = 0
            ## make plot of predicted vs empirical ownership profile
            #aggregateOutTable = tableDict.fromTextFile(aggregateOut, width = 20, prec = 10)
            #aggregateOutTable.keep(['age', 'owner'])
            #aggregateOutTable.rename('owner', 'thOwnership')
            #empOwnershipTable = tableDict.fromTextFile(empOwnershipFile, width = 20, prec = 10)
            #empOwnershipTable.rename('PrOwnership', 'empOwnership')
            #ownershipTable = aggregateOutTable.merged(empOwnershipTable, 'age')
            #ownershipTable.drop('_merge')
            #yVars = ['thOwnership', 'empOwnership']
            ## add the individual simulations
            #for profile in [ '1', '2', '3' ]:
                #profileOwnershipFile = os.path.join(output_dir, 'simulation_' + profile + '.out')
                #if not os.path.exists(profileOwnershipFile): continue
                #profileOwnershipTable = tableDict.fromTextFile(profileOwnershipFile, width = 20, prec = 10)
                #profileOwnershipTable.keep(['age', 'owner'])
                #profileOwnershipTable.rename('owner', 'thOwnership_' + profile)
                #ownershipTable.merge(profileOwnershipTable, 'age')
                #ownershipTable.drop('_merge')
                #yVars.append('thOwnership_' + profile)
            #f = open(ownershipTableFile, 'w')
            #print >> f, ownershipTable
            #f.close()
            #try:
                #plotSimulation(table = ownershipTableFile, xVar = 'age', yVars = yVars, yVarRange = (0., 1.), figureFile = os.path.join(self.output_dir, 'ownership.png'), verb = 'CRITICAL')
            #except:
                #logger.debug('couldnt make ownershipTableFile')

            #try:
                #self.execution.exitcode = plotOwnerProfiles(path2input = os.path.join(os.getcwd(), 'localBaseDir', 'input'),
                                                        #path2output = output_dir, simuFileName = 'aggregate.out')
            #except:
                #logger.debug('couldnt make ownerprofile plot. plotOwnerProfiles crashed, couldnt even set exitcode. ')
            # make plot of life-cycle simulation (all variables)
            try:
                makeAggregatePlot(self.output_dir)
#                plotSimulation(table = os.path.join(output_dir, 'aggregate.out'), xVar = 'age', yVars = ['wealth', 'theta1', 'theta2', 'theta3', 'theta4', 'theta5', 'cons', 'income'], figureFile = os.path.join(self.output_dir, 'aggregate.png'), verb = 'CRITICAL' )
            except:
                logger.debug('coulndt make aggregate.out plot')
            #if os.path.exists('ownershipThreshold_1.out'):
                #plotSimulation(path = os.path.join(output_dir, 'ownershipThreshold_1.out'), xVar = 'age', yVars = [ 'Yst1', 'Yst4' ], figureFile = os.path.join(self.output_dir, 'ownershipThreshold_1.eps'), verb = 'CRITICAL' )
                #plotSimulation(path = os.path.join(output_dir, 'ownershipThreshold_1.out'), xVar = 'age', yVars = [ 'yst1', 'yst4' ], figureFile = os.path.join(self.output_dir, 'normownershipThreshold_1.eps'), verb = 'CRITICAL' )
        else:
            # no `simulation.out` found, signal error
            self.execution.exitcode = 2

class housingApppotApplication(housingApplication, gc3libs.application.apppot.AppPotApplication):
    _invalid_chars = re.compile(r'[^_a-zA-Z0-9]+', re.X)
    def __init__(self, executable, arguments, inputs, outputs, output_dir, **extra_args):
#        extra_args.setdefault('requested_walltime', 2*hours) # unnecessary.. gc3pie automatically sets default to 8
        gc3libs.application.apppot.AppPotApplication.__init__(self, executable, arguments, inputs, outputs, output_dir, **extra_args)



#! /usr/bin/env python
#
#   gipyrad.py -- Front-end script for submitting multiple `IPyRAD` jobs.
#
#   Copyright (C) 2017 2018  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `PyRAD` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gipyrad --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

__docformat__ = 'reStructuredText'
# summary of user-visible changes
__changelog__ = """
  2017-11-27:
  * initial working version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gipyrad
    gipyrad.GipyradScript().run()

#from pkg_resources import Requirement, resource_filename
import os
import posix

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, existing_directory
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds, GiB
from gc3libs.workflow import RetryableTask
import gc3libs.url

# Default values
FASTQ_FORMATS = ["fastq","fastq.gz"]
DATA_PATH="/data"
IPYRAD_STEPS="1234567"
IPYRAD_PARAMFILE_NAME="params.txt.modified"
IPYRAD_COMMAND="sudo docker run --rm -v $PWD:{0} smaffiol/ipyrad -p {0}/{1} -s{2}"
IPYRAD_PARAMFILE_PATTERN="sorted_fastq_path"
IPYRAD_PARAMFILE_REPLACEMENT_STRING="{0}/*.fastq.gz".format(DATA_PATH)

## Utility methods

def get_valid_input_pair(input_folder):
    """
    scans input folder [non recursively]
    for each valid compressed fastq file [fastq.gz]
    search for a pair of type [R1,R2].
    Search is done at filename level.
    """
    R1list = [infile.split('R1')[0] for infile in os.listdir(input_folder) \
              if infile.endswith('R1.fastq.gz')]

    input_list = [(os.path.join(input_folder,
                                "{0}R1.fastq.gz".format(infile.split('R2')[0])),
                   os.path.join(input_folder,
                                "{0}R2.fastq.gz".format(infile.split('R2')[0]))) for infile in os.listdir(input_folder) if infile.endswith('R2.fastq.gz') and infile.split('R2')[0] in R1list]

    return input_list

def prepare_ipyrad_param_file(pyrad_param_file):
    """
    Replace input data references and prepare for docker execution
    change [4] to reflect docker invokation:
    /data/*.fastq.gz
    """
    with open(IPYRAD_PARAMFILE_NAME,'w+') as wd:
        with open(pyrad_param_file,'r') as fd:
            for line in fd:
                if IPYRAD_PARAMFILE_PATTERN in line:
                    wd.write(IPYRAD_PARAMFILE_REPLACEMENT_STRING)
                else:
                    wd.write(line)
    return wd.name

## custom application class
class GipyradApplication(Application):
    """
    Fetches execution wrapper, input file and checks
    whether optional arguments have been passed.
    """

    application_name = 'gipyrad'

    def __init__(self, input_files, dockerimage, paramsfile, ipyradsteps, **extra_args):
        """
        iPyrad is executed within docker container
        """

        inputs = []
        outputs = []

        inputs.append(paramsfile)
        for input_file in input_files:
            inputs.append(input_file)

        Application.__init__(
            self,
            arguments = IPYRAD_COMMAND.format(DATA_PATH,
                                              IPYRAD_PARAMFILE_NAME,
                                              ipyradsteps),
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'gipyrad.log',
            join=True,
            **extra_args)

## main script class

class GipyradScript(SessionBasedScript):
    """
Scan the specified INPUT directories for valid fastq files and
submit a job for each valid pair found; job progress is
monitored and, when a job is done, its output files are retrieved back
into the simulation directory itself.

The ``gipyrad`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gipyrad`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GipyradApplication,
            stats_only_for = GipyradApplication,
            )

    def setup_args(self):
        self.add_param('input_folder', type=existing_directory,
                       help="Path to folder containing fastq files.")

    def setup_options(self):
        self.add_param("-D", "--dockerimage", metavar="STRING",
                       dest="dockerimage", default="smaffiol/ipyrad",
                       help="Docker image to be used to execute ipyrad." \
                       " Default: %(default)s")

        self.add_param("-S", "--ipyrad-steps", metavar="STRING",
                       dest="ipyradsteps", default=IPYRAD_STEPS,
                       help="ipyrad Set of assembly steps to perform." \
                       " e.g., -S 123. Default: %(default)s")

        self.add_param("-P", "--params", metavar="PATH",
                       dest="paramsfile", default='./params.txt',
                       help="path to ipyrad params file for Assembly." \
                       " Default: %(default)s")

    def new_tasks(self, extra):

        paramsfile = prepare_ipyrad_param_file(self.params.paramsfile)
        tasks = []

        for input_file in get_valid_input_pair(self.params.input_folder):

            # extract jobname from the 1st file of the input_file pair
            jobname = "%s" % os.path.basename(input_file[0]).split(".fastq.gz")[0]

            extra_args = extra.copy()
            extra_args['jobname'] = jobname

            # FIXME: ignore SessionBasedScript feature of customizing
            # output folder
            extra_args['output_dir'] = self.params.output

            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

            self.log.info("Creating Task: [{0}]".format(jobname))

            tasks.append(GipyradApplication(
                input_file,
                self.params.dockerimage,
                paramsfile,
                self.params.ipyradsteps,
                **extra_args
            ))

        return tasks



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011, 2014, 2015, 2019,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'


import re

from gc3libs import log
from gc3libs.backends.transport import LocalTransport


class FakeTransport(LocalTransport):

    """
    This class allows you to override responses to specific commands.

    The `execute_command` of `FakeTransport` instead of executing a
    command will look into the `self.expected_answer` dictionary and
    find a proper triple `(retcode, output, error)`. If not found, the
    command will be executed by the parent (`LocalTransport`)

    To use this class you need to fill the `self.expected_answer` with
    commands (only the *binary*, not the whole command line) as keys,
    and triples (retcode, output, error) as values. For example:

    >>> t = FakeTransport()
    >>> t.expected_answer['echo'] = (0, 'GC3pie is wonderful', '')
    >>> t.execute_command("echo 'GC3pie is wonderful'") \
        == (0, 'GC3pie is wonderful', '')
    True

    FIXME: The whole logic in this class is flawed.  GC3Pie does many
    layers of quoting and wrapping commands, to remove which we would
    need a complete and correct ``sh`` parser (no, the `shlex` module
    is not enough).  We should be passing argv-style arrays to the
    `Transport.execute_command` module and check that instead; see
    https://github.com/uzh/gc3pie/issues/285&q=argv
    """

    def __init__(self, expected_answer={}):
        LocalTransport.__init__(self)
        self.expected_answer = expected_answer

    _COMMAND_RE = re.compile(
        # match start of the line or any shell metacharacter that can
        # introduce a new command (XXX: this is quite approximate,
        # e.g. ``)`` will only be a command-introducing syntax in the
        # ``case ... esac`` body).
        r"(^|;|&|`|\(|\))"
        # optional white space
        r" \s*"
        # optional quotes, including doubly-nested quotes
        r" ('|'\\''" r'|")?'
        # the actual command (XXX: strictly speaking this is incorrect
        # as ``foo"bar"`` is a valid shell word that evaluates to
        # ``foobar`` but again there's no way we can really parse
        # ``sh`` syntax with regexps...)
        " (?P<cmd>[a-z0-9_+-]+)",
        re.VERBOSE | re.IGNORECASE)

    def execute_command(self, cmdline):
        """
        Scan the given command-line and return a predefined result if
        *any* word in command position matches one of the keys in the
        `expected_answer` argument to the class constructor.

        Note that the parsing of command-line is based on regular
        expressions and is thus only an approximation at ``sh``
        syntax.  It will *certainly* fail on some command-lines, but
        there is no way around this short of writing a complete ``sh``
        parser just for this function.  (And no, Python's module
        `shlex` will not do the job -- been there, done that.)
        """

        log.debug("scanning command-line <<<%s>>>", cmdline)

        for match in self._COMMAND_RE.finditer(cmdline):
            cmd = match.group("cmd")
            if cmd in self.expected_answer:
                reply = self.expected_answer[cmd]
                log.debug("returning programmed reply for '%s': %s", cmd, reply)
                return reply

        # if everything else failed, do run the command-line ...
        return LocalTransport.execute_command(self, cmdline)


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="faketransport",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python

"""
Fake running applications, only useful for testing.
"""
# Copyright (C) 2009-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'


# stdlib imports
from random import random

# GC3Pie imports
import gc3libs
import gc3libs.exceptions
from gc3libs import log, Run
from gc3libs.utils import same_docstring_as
from gc3libs.backends import LRMS
from gc3libs.quantity import Memory


NORMAL_TRANSITION_GRAPH = {
    Run.State.SUBMITTED:   {1.0: Run.State.RUNNING},
    Run.State.RUNNING:     {1.0: Run.State.TERMINATING},
    Run.State.TERMINATING: {1.0: Run.State.TERMINATED},
}


class NoOpLrms(LRMS):
    """
    Simulate execution of an `Application`:class instance.

    Upon every invocation of `update_job_state`:meth: the application
    status is advanced to the next state (according to the normal
    progression SUBMITTED -> RUNNING -> TERMINATING).

    This progression can be altered by assigning a different
    transition graph to attribute `transition_graph` on an instance.
    The transition graph has a two-level structure:

    * keys are task execution states (e.g., `Run.State.SUBMITTED`)

    * values are dictionaries, mapping a probability (i.e., a floating
      point number between 0.0 and 1.0) to a new state.  All
      probabilities should sum to a number less then, or equal to, 1.0
      -- but this condition is not checked or enforced.  Likewise, it
      is not checked nor enforced that the new state is a valid target
      state given the source.

    Every invocation of :meth:`update_job_state` results in the task
    execution state possibly changing to one of the target states,
    according to the given transition probabilities.

    For example, the following transition graph specifies that a job
    in state ``SUBMITTED`` can change to ``RUNNING`` with 80%
    probability (and with 20% stay in ``SUBMITTED`` state); a job in
    state ``RUNNING`` has a 50% chance of transitioning to
    ``TERMINATING``, 10% chance of being ``STOPPED`` and 40% chance of
    staying in state ``RUNNING``; and a job in ``STOPPED`` state stays
    in ``STOPPED`` state forever::

    | transition_graph = {
    |   Run.State.SUBMITTED = {
    |     0.80: Run.State.RUNNING,
    |   },
    |   Run.State.RUNNING = {
    |     0.50: Run.State.TERMINATING,
    |     0.10: Run.State.STOPPED,
    |     0.40: Run.State.RUNNING, # implcit, could be omitted
    |   },
    |   Run.State.STOPPED = {
    |     1.00: Run.State.STOPPED,
    |   },
    | }

    All parameters taken by the base class `LRMS`:class: are
    understood by this class constructor, but they are actually
    ignored.
    """

    def __init__(self, name,
                 # these parameters are inherited from the `LRMS` class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 auth=None,
                 **extra_args):

        # init base class
        LRMS.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth, **extra_args)

        self.transition_graph = NORMAL_TRANSITION_GRAPH

        # initial resource status;
        # use `max_cores` as the max number of processes to allow
        self.user_queued = 0
        self.free_slots = self.max_cores
        self.queued = 0
        self.user_run = 0
        self.available_memory = self.max_cores * self.max_memory_per_core

    @same_docstring_as(LRMS.cancel_job)
    def cancel_job(self, app):
        """
        Simulate stopping a job's process.

        Effectively, this just returns any resources accounted to the
        job into the free pool.
        """
        if app.execution.state == Run.State.RUNNING:
            self.user_run -= 1
        elif app.execution.state == Run.State.SUBMITTED:
            self.user_queued -= 1
            self.queued -= 1
        self.free_slots += app.requested_cores
        if app.requested_memory:
            self.available_memory += app.requested_memory

    def close(self):
        """This is a no-op for this backend."""
        pass

    def free(self, app):
        """This is a no-op for this backend."""
        pass

    @same_docstring_as(LRMS.get_resource_status)
    def get_resource_status(self):
        # if we have been doing our own book-keeping well, then
        # there's no resource status to update
        self.updated = True
        return self

    @same_docstring_as(LRMS.get_results)
    def get_results(self, app, download_dir,
                    overwrite=False, changed_only=True):
        if app.outputs:
            raise gc3libs.exceptions.DataStagingError(
                "Retrieval of output files is not supported"
                " in the NoOp backend.")
        return

    def update_job_state(self, app):
        """
        Advance `app`'s status to the next one
        in the normal execution graph.
        """
        log.debug("No-Op backend updating state of Task %s ...", app)
        transitions = self.transition_graph[app.execution.state]
        log.debug("Task %s transitions: %s.", app, ", ".join([
            ("with probability %g to state %s" % (prob, state))
            for prob, state in list(transitions.items()) if prob > 0
        ]))
        dice = random()
        # log.debug("Rolled dice, got %g result", dice)
        for prob, to_state in sorted(transitions.items()):
            if dice < prob:
                log.debug(
                    "Task %s transitions to state '%s'", app, to_state)
                # update resource state based on old and new app state
                if app.execution.state == Run.State.SUBMITTED:
                    self.queued -= 1
                    self.user_queued -= 1
                if app.execution.state == Run.State.RUNNING:
                    self.user_run -= 1
                if to_state == Run.State.RUNNING:
                    self.user_run += 1
                if to_state == Run.State.TERMINATING:
                    self.free_slots += app.requested_cores
                    if app.requested_memory:
                        self.available_memory += app.requested_memory
                # set the new app state
                app.execution.state = to_state
                break
            else:
                dice -= prob
        return app.execution.state

    def submit_job(self, app):
        """
        Transition `app`'s status to `Run.State.SUBMITTED` if possible.

        Note that this method still checks that `app`'s requirements
        are compatible with what this resource was instanciated with,
        and that conversely the resource still has enough free
        cores/memory/etc to host a new application.  So, submission to
        a No-Op resource may still fail!
        """
        if app.requested_cores > self.free_slots:
            raise gc3libs.exceptions.MaximumCapacityReached(
                "Resource %s does not have enough free cores:"
                " %s requested, but %s available."
                " Increase 'max_cores' to raise."
                % (self.name, app.requested_cores, self.free_slots))

        if (app.requested_memory and
                 self.available_memory < app.requested_memory):
            raise gc3libs.exceptions.MaximumCapacityReached(
                "Resource %s does not have enough available memory:"
                " %s requested, but only %s available."
                % (self.name,
                   app.requested_memory.to_str('%g%s', unit=Memory.MB),
                   app.available_memory.to_str('%g%s', unit=Memory.MB),)
            )

        log.debug("Faking execution of command '%s' ...",
                  " ".join(app.arguments))

        # Update application and current resources
        app.execution.lrms_jobid = id(app)
        self.free_slots -= app.requested_cores
        if app.requested_memory:
            self.available_memory -= app.requested_memory
        self.queued += 1
        self.user_queued += 1

        return app

    def peek(self, app, remote_filename, local_file, offset=0, size=None):
        """Not supported on this backend."""
        raise NotImplementedError(
            "The `peek` operation is not supported"
            " by the `NoOp` backend.")

    def validate_data(self, data_file_list=[]):
        """
        Return `False` if any of the URLs in `data_file_list` cannot
        be handled by this backend.

        The `noop`:mod: backend can not do *any* kind of I/O, so this
        method will only return `True` if the supplied list of files
        is empty.
        """
        for url in data_file_list:
            if url.scheme:
                return False
        return True


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="__init__",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gndn.py -- Front-end script for running fitModel.R
#   function over a defined Data set and a given model.
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gndn.py --help`` for program usage
instructions.

Input parameters consists of:

...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-02-17:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gndn
    gndn.GndnScript().run()

import os
import sys
import time
import tempfile

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_CORES = 4
DEFAULT_MEMORY = Memory(7,GB)
DEFAULT_WALLTIME = Duration(200,hours)

## custom application class
class GndnApplication(Application):
    """
    Custom class to wrap the execution of fitModel.R script
    The system is agnostic of what exactly and how exactly fitModel.R is
    going to be executed.
    Just read the content of `command.txt` file in the root of the input folder
    and executed it.
    A convenient wrapper script is provided to facilitate how the command file
    is executed on the remote end.
    """
    application_name = 'gndn'

    def __init__(self, input_folder, **extra_args):

        self.output_dir = extra_args['output_dir']
        self.input_folder = input_folder

        inputs = dict()
        outputs = dict()

        gndn_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gndn_wrapper.sh")

        inputs[gndn_wrapper_sh] = os.path.basename(gndn_wrapper_sh)

        inputs[input_folder] = "%s/" % os.path.basename(input_folder)
        outputs[os.path.join(os.path.basename(input_folder),"results")] = "results/"

        arguments = "./%s %s" % (inputs[gndn_wrapper_sh],inputs[input_folder])

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gndn.log',
            join=True,
            executables = "./%s" % os.path.basename(input_folder),
            **extra_args)

    def terminated(self):
        """
        Move results into original results folder
        """
        gc3libs.log.info("Application terminated with exit code %s" % self.execution.exitcode)
        for result in os.listdir(os.path.join(self.output_dir,"results/")):
            shutil.move(os.path.join(self.output_dir,"results/",result),
                        os.path.join(self.input_folder,"results",os.path.basename(result)))
        # Cleanup
        os.removedirs(os.path.join(self.output_dir,"results/"))

class GndnScript(SessionBasedScript):
    """
    The script takes as input a comma separated list of input folders.
    Each input folder contains all the information to run a fitModel with a given
    Dataset and a defined model.
    The input folder will contain a command file that will mimic exactly the invocation
    of fitModel.R for the given input folder as executed on a local computer.
    For each input folder, `gdnd` creates an instance of GndnApplication.

    The ``gndn`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gndn`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GndnApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GndnApplication,
            )

    def parse_args(self):
        """
        Check that each element in args is at least a alid folder
        """
        for folder_name in self.params.args:
            if not os.path.isdir(folder_name):
                gc3libs.log.error(
                    "Invalid input folder: {folder_name}."
                    " Removing it from input list"
                    .format(folder_name=folder_name))
                self.params.args.remove(folder_name)

    def new_tasks(self, extra):
        """
        For each input folder, create an instance of GndnApplication
        """
        tasks = []

        for input_folder in self.params.args:

            # extract root folder name to be used as jobname
            jobname = os.path.basename(input_folder)

            extra_args = extra.copy()

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', 'run_%s' % jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', 'run_%s' % jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', 'run_%s' % jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', 'run_%s' % jobname)

            tasks.append(GndnApplication(
                os.path.abspath(input_folder),
                **extra_args))

        return tasks



#! /usr/bin/env python
#
"""
Test class `Task`:class:.
"""
# Copyright (C) 2011, 2012, 2018, 2019,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'


## imports
import pytest

from gc3libs import Run, Task


from gc3libs.testing.helpers import SuccessfulApp, UnsuccessfulApp, temporary_core


## tests

def test_task_progress():
    with temporary_core() as core:
        task = SuccessfulApp()
        task.attach(core)
        task.submit()
        assert task.execution.state == Run.State.SUBMITTED

        # run until terminated
        while task.execution.state != Run.State.TERMINATED:
            task.progress()
        assert task.execution.state == Run.State.TERMINATED


def test_task_state_handlers():
    report = {
        'submitted': False,
        'terminating': False,
        'terminated': False,
    }

    class HandlerTestApp(SuccessfulApp):
        def __init__(self, report):
            super(HandlerTestApp, self).__init__()
            self.report = report
        def submitted(self):
            self.report['submitted'] = True
            print("Submitted!")
        def terminating(self):
            self.report['terminating'] = True
            print("Terminating!")
        def terminated(self):
            print("Terminated!")
            self.report['terminated'] = True

    with temporary_core() as core:
        task = HandlerTestApp(report)
        task.attach(core)
        task.submit()
        assert report['submitted']

        # run until terminated
        while task.execution.state != Run.State.TERMINATED:
            task.progress()
        assert report['terminating']
        assert report['terminated']


def test_task_redo1():
    """Test correct use of `Task.redo()`"""
    with temporary_core() as core:
        task = SuccessfulApp()
        task.attach(core)
        task.submit()
        assert task.execution.state == Run.State.SUBMITTED

        # run until terminated
        while task.execution.state != Run.State.TERMINATED:
            task.progress()
        assert task.execution.state == Run.State.TERMINATED

        # no do it all over again
        task.redo()
        assert task.execution.state == Run.State.NEW

        task.progress()
        assert task.execution.state != Run.State.NEW
        assert task.execution.state in [Run.State.SUBMITTED, Run.State.RUNNING]


def test_task_redo2():
    """Test that `.redo()` raises if called on a Task that is not TERMINATED."""
    with temporary_core() as core:
        task = SuccessfulApp()
        task.attach(core)
        task.submit()
        assert task.execution.state == Run.State.SUBMITTED

        # cannot redo a task that is not yet terminated
        with pytest.raises(AssertionError):
            task.redo()
            pytest.fail("`Task.redo()` succeeded on task not yet finished")


def test_task_redo3():
    """Test that `.redo()` is a no-op if the Task is still NEW."""
    with temporary_core() as core:
        task = SuccessfulApp()
        task.attach(core)
        task.redo()


# main: run tests

if "__main__" == __name__:
    import pytest
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
#   gnlp.py -- Front-end script for evaluating CoreNLP Sentiments
#   over a large dataset.
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple CoreNLP jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gnlp.py --help`` for program usage
instructions.

Input parameters consists of:
:param str input file: Path to an .xml file containing input data with structure
<ROWSET>
<ROW>
<FIELD1>
<PostId>
<ThreadID>
<UserID>
<TimeStamp>
<Upvotes>
<Downvotes>
<Flagged>
<Approved>
<Deleted>
<Replies>
<ReplyTo>
<Content>
</ROW>
<ROW>
...
</ROWSET>
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2014-09-05:
  * Added parsing of output file
  * Merging all results into single XML
  * Added <Sentiment> tag for results
  2014-08-14:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gnlp
    gnlp.GnlpScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
from xml.etree import cElementTree as ET
from xml.etree.ElementTree import Element, SubElement, Comment, tostring

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


XML_HEADER = """<?xml version="1.0"?>
<ROWSET>
"""
XML_FOOTER = "</ROWSET>"


## custom application class
class GnlpApplication(Application):
    """
    Custom class to wrap the execution of the CoreNLP java script.
    """
    application_name = 'corenlp'

    def __init__(self, input_data, **extra_args):

        # setup input references
        inputs = dict()

        inputs[input_data] = "./input.txt"

        # adding wrapper main script
        gnlp_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gnlp_wrapper.py")

        inputs[gnlp_wrapper_sh] = "./wrapper.py"

        arguments = "./wrapper.py ./input.txt ./output.txt"

        outputs = dict()
        outputs['./output.txt'] = extra_args['output_file']

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gnlp.log',
            stderr = 'gnlp.err',
            executables = "./wrapper.py",
            **extra_args)


class GnlpScript(SessionBasedScript):
    """
    Splits input .xml file into smaller chunks, each of them of size
    'self.params.chunk_size'.
    Then it submits one execution for each of the created chunked files.

    The ``gnlp`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gnlp`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gnlp``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GnlpApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GnlpApplication,
            )

    def setup_options(self):
        self.add_param("-k", "--chunk", metavar="[NUM]",
                       dest="chunk_size", default="1000",
                       help="How to split the .XML input data set. "
                       "Default: 1000")

        self.add_param("--result-file", metavar="[STRING]",
                       dest="result_file", default='result.xml',
                       help="Name of the result file generated as the aggregation"
                       " of all results from each chunked execution. "
                       "Default: result.xml")

    def setup_args(self):

        self.add_param('input_data', type=str,
                       help="Input data full path name.")

    def parse_args(self):
        """
        Check presence of input file.
        """

        # check args:
        if not os.path.isfile(self.params.input_data):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid path to input data: '%s'. File not found"
                % self.params.input_data)

        # Verify that 'self.params.chunk_size' is int
        int(self.params.chunk_size)
        self.params.chunk_size = int(self.params.chunk_size)

    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new GcgpsTask
        """
        tasks = []
        last_index = 0

        for (input_file, index_chunk) in self._generate_chunked_files_and_list(self.params.input_data,
                                                                              self.params.chunk_size):

            jobname = "gnlp-%d-%d" % (last_index,index_chunk)

            extra_args = extra.copy()

            extra_args['index_chunk'] = str(index_chunk)

            extra_args['jobname'] = jobname

            # extra_args['output_file'] = 'result.xml'
            extra_args['output_file'] = self.params.result_file

            extra_args['output_dir'] = os.path.abspath(self.params.output)
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

            self.log.debug("Creating Application for index : %d - %d" %
                           (last_index,index_chunk))

            tasks.append(GnlpApplication(
                    input_file,
                    **extra_args))

            last_index = index_chunk


        return tasks

    def after_main_loop(self):
        """
        Merge all result files together
        Then clean up tmp files.
        Format of output file:
        FiledID@<Sentiments>..</Sentiments>
        e.g.
        296@<Sentiments>Neutral</Sentiments>
        This means that the <Sentiments> tag should be added to the XML element
        corresponding to <Filed1>296</Field1>
        """

        # Open Session result file
        try:
            with open(self.params.result_file,'w+') as fd:
                # Write header
                fd.write(XML_HEADER)

                for task in self.session:
                    if isinstance(task,GnlpApplication) and task.execution.returncode == 0:
                        with open(os.path.join(task.output_dir,task.output_file),'r') as fin:
                            for line in fin:
                                if line.strip() in ['<?xml version="1.0"?>','<ROWSET>','</ROWSET>']:
                                    continue
                                fd.write(line)

                fd.write(XML_FOOTER)
        except OSError, osx:
            gc3libs.log.error("Failed while merging results. Error %s", str(osx))

    def _generate_chunked_files_and_list(self, file_to_chunk, chunk_size=1000):
        """
        Takes a file_name as input and a defined chunk size
        ( uses 1000 as default )
        returns a list of filenames 1 for each chunk created and a corresponding
        index reference
        e.g. ('/tmp/chunk2800.xml,2800) for the chunk segment that goes
        from 2800 to (2800 + chunk_size)
        """

        index = 0
        chunk = []
        fout = None
        failure = False
        chunk_files_dir = os.path.join(self.session.path,"tmp")

        # creating 'chunk_files_dir'
        if not(os.path.isdir(chunk_files_dir)):
            try:
                os.mkdir(chunk_files_dir)
            except OSError, osx:
                gc3libs.log.error("Failed while creating tmp folder %s. " % chunk_files_dir +
                                  "Error %s." % str(osx) +
                                  "Using default '/tmp'")
                chunk_files_dir = "/tmp"

        try:

            with open(file_to_chunk,'r') as fin:
                for line in fin:
                    # Ignore header and ROWSET
                    if line.strip() in ['<?xml version="1.0"?>','<ROWSET>','']:
                        # Ignore and continue
                        continue

                    if line.strip() == "<row>":
                        if index % chunk_size == 0:
                            if fout:
                                # Close existing file and create a new one
                                fout.write(XML_FOOTER)
                                fout.close()
                                chunk.append((fout.name, index))
                            # Create new chunk file
                            fout = open(os.path.join(chunk_files_dir,"input-%d" % index),"w")
                            fout.write(XML_HEADER)
                        index += 1
                    if not fout:
                        fout = open(os.path.join(chunk_files_dir,"input-%d" % index),"w")
                    fout.write(line)

            # Just close the current fout file if needed
            fout.close()
            chunk.append((fout.name, index))

        except OSError, osx:
            gc3libs.log.critical("Failed while creating chunk files." +
                                 "Error %s", (str(osx)))
            failure = True
        finally:
            if failure:
                # remove all tmp file created
                gc3libs.log.info("Could not generate full chunk list. "
                                 "Removing all existing tmp files... ")
                for (cfile, index) in chunk:
                    try:
                        os.remove(cfile)
                    except OSError, osx:
                        gc3libs.log.error("Failed while removing " +
                                          "tmp file %s. " +
                                          "Message %s" % osx.message)

        return chunk



#! /usr/bin/env python

"""
Generic Python programming utility functions.

This module collects general utility functions, not specifically
related to GC3Libs.  A good rule of thumb for determining if a
function or class belongs in here is the following: place a function
or class in this module if you could copy its code into the
sources of a different project and it would not stop working.
"""

# Copyright (C) 2009-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import next
from builtins import range
from builtins import object
__docformat__ = 'reStructuredText'


from codecs import decode
from collections import defaultdict, deque
try:
    # Python 3
    from collections.abc import Mapping, MutableMapping
except ImportError:
    from collections import Mapping, MutableMapping
import contextlib
import functools
from .compat._functools import total_ordering
import locale
import os
import os.path
if bytes == str:
    # Python2's `os.sysconf()` cannot handle sysconf variable names as
    # Unicode strings, so we need to cast them to byte-strings
    def os_sysconf(name):
        return os.sysconf(str(name))
else:
    # Python3
    from os import sysconf as os_sysconf
import random
import re
import shutil
import sys
import tempfile
import time
import io as StringIO


from gc3libs.compat._collections import OrderedDict
from gc3libs.compat._functools import total_ordering
import lockfile

import gc3libs
import gc3libs.exceptions
import gc3libs.debug
from gc3libs.quantity import Duration, Memory


if bytes == str:
    # Python 2
    text_str = unicode
else:
    # Python 3+
    text_str = str


# This fixes an issue with Python 2.4, which does not have
# `shutl.WindowsError`
try:
    WindowsError = shutil.WindowsError
except AttributeError:
    class NeverUsedException(Exception):

        """this exception should never be raised"""
    WindowsError = NeverUsedException


# map suffix to pow multipliers. This is 2.6 compatible.
BYTE_SUFFIX_TO_POW = dict(((l, i)) for i, l in enumerate('kmgtpezy', start=1))
# 2.7 +:  mapped = {l: i for i, l in enumerate('kmgtpezy', start=1)}


def backup(path):
    """
    Rename the filesystem entry at `path` by appending a unique
    numerical suffix; return new name.

    For example,

    1. create a test file:

      >>> import tempfile
      >>> path = tempfile.mkstemp()[1]

    2. then make a backup of it; the backup will end in ``.~1~``:

      >>> path1 = backup(path)
      >>> os.path.exists(path + '.~1~')
      True

    3. re-create the file, and make a second backup: this time the
    file will be renamed with a ``.~2~`` extension:

      >>> open(path, 'w').close()
      >>> path2 = backup(path)
      >>> os.path.exists(path + '.~2~')
      True

    cleaning up tests

      >>> os.remove(path+'.~1~')
      >>> os.remove(path+'.~2~')

    """
    parent_dir = (os.path.dirname(path) or os.getcwd())
    prefix = os.path.basename(path) + '.~'
    p = len(prefix)
    suffix = 1
    for name in [entry for entry in os.listdir(parent_dir)
                 if (entry.startswith(prefix) and entry.endswith('~'))]:
        try:
            n = int(name[p:-1])
            suffix = max(suffix, n + 1)
        except ValueError:
            # ignore non-numeric suffixes
            pass
    new_path = "%s.~%d~" % (path, suffix)
    os.rename(path, new_path)
    return new_path


def basename_sans(path):
    """
    Return base name without the extension.

    This behaves exactly like :func:`os.path.basename` except that the
    last few characters, up to the rightmost dot, are removed as
    well::

      >>> basename_sans('/tmp/foo.txt') == 'foo'
      True

      >>> basename_sans('bar.txt') == 'bar'
      True

    If there is no dot in the file name, no "extension" is chopped
    off::

      >>> basename_sans('baz') == 'baz'
      True

    If there are several dots in the file name, only the last one and
    trailing characters are removed::

      >>> basename_sans('foo.bar.baz') == 'foo.bar'
      True

    Leading directory components are chopped off in any case::

      >>> basename_sans('/tmp/foo.bar.baz') == 'foo.bar'
      True

      >>> basename_sans('/tmp/foo') == 'foo'
      True
    """
    return os.path.splitext(os.path.basename(path))[0]


def cache_for(lapse):
    """
    Cache the result of a (nullary) method invocation for a given
    amount of time. Use as a decorator on object methods whose results
    are to be cached.

    Store the result of the first invocation of the decorated
    method; if another invocation happens before `lapse` seconds
    have passed, return the cached value instead of calling the real
    function again.  If a new call happens after the grace period has
    expired, call the real function and store the result in the cache.

    **Note:** Do not use with methods that take keyword arguments, as
    they will be discarded! In addition, arguments are compared to
    elements in the cache by *identity*, so that invoking the same
    method with equal but distinct object will result in two separate
    copies of the result being computed and stored in the cache.

    Cache results and timestamps are stored into the objects'
    `_cache_value` and `_cache_last_updated` attributes, so the caches
    are destroyed with the object when it goes out of scope.

    The working of the cached method can be demonstrated by the
    following simple code::

        >>> class X(object):
        ...     def __init__(self):
        ...         self.times = 0
        ...     @cache_for(2)
        ...     def foo(self):
        ...             self.times += 1
        ...             return self.times
        >>> x = X()
        >>> x.foo()
        1
        >>> x.foo()
        1
        >>> time.sleep(3)
        >>> x.foo()
        2

    """
    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(obj, *args):
            now = time.time()
            key = (fn, tuple(id(arg) for arg in args))
            try:
                update = ((now - obj._cache_last_updated[key]) > lapse)
            except AttributeError:
                obj._cache_last_updated = defaultdict(lambda: 0)
                obj._cache_value = dict()
                update = True
            if update:
                obj._cache_value[key] = fn(obj, *args)
                obj._cache_last_updated[key] = now
            # gc3libs.log.debug("%s(%s, ...): Using cached value '%s'",
            #                  fn.__name__, obj, obj._cache_value[key])
            return obj._cache_value[key]
        return wrapper
    return decorator


def cat(*args, **extra_args):
    """
    Concatenate the contents of all `args` into `output`.  Both
    `output` and each of the `args` can be a file-like object or a
    string (indicating the path of a file to open).

    If `append` is `True`, then `output` is opened in append-only
    mode; otherwise it is overwritten.
    """
    output = extra_args.get('output', sys.stdout)
    append = extra_args.get('append', True)
    # ensure `output` is a file-like object, opened in write-mode
    try:
        output.write('')
    except:
        output = open(output, ('a' if append else 'w'))
    for arg in args:
        # ensure `arg` is a file-like object, opened in read-mode
        try:
            arg.read(0)
        except:
            arg = open(arg, 'r')
        for line in arg:
            output.write(line)


def check_file_access(path, mode, exception=RuntimeError, isdir=False):
    """
    Test for access to a path; if access is not granted, raise an
    instance of `exception` with an appropriate error message.
    This is a frontend to `os.access`:func:, which see for exact
    semantics and the meaning of `path` and `mode`.

    :param path: Filesystem path to test.
    :param mode: See `os.access`:func:
    :param exception: Class of exception to raise if test fails.
    :param isdir: If `True` then also test that `path` points to a directory.

    If the test succeeds, `True` is returned::

      >>> check_file_access('/bin/cat', os.F_OK)
      True
      >>> check_file_access('/bin/cat', os.R_OK)
      True
      >>> check_file_access('/bin/cat', os.X_OK)
      True
      >>> check_file_access('/tmp', os.X_OK)
      True

    However, if the test fails, then an exception is raised::

      >>> check_file_access('/bin/cat', os.W_OK)
      Traceback (most recent call last):
        ...
      RuntimeError: Cannot write to file '/bin/cat'.

    If the optional argument `isdir` is `True`, then additionally test
    that `path` points to a directory inode::

      >>> check_file_access('/tmp', os.F_OK, isdir=True)
      True

      >>> check_file_access('/bin/cat', os.F_OK, isdir=True)
      Traceback (most recent call last):
        ...
      RuntimeError: Expected '/bin/cat' to be a directory, but it's not.
    """
    what = ("directory" if isdir else "file")
    if not os.access(path, os.F_OK):
        raise exception("Cannot access %s '%s'." % (what, path))
    if isdir and not os.path.isdir(path):
        raise exception(
            "Expected '%s' to be a directory, but it's not." % path)
    if (mode & os.R_OK) and not os.access(path, os.R_OK):
        raise exception("Cannot read %s '%s'." % (what, path))
    if (mode & os.W_OK) and not os.access(path, os.W_OK):
        raise exception("Cannot write to %s '%s'." % (what, path))
    if (mode & os.X_OK) and not os.access(path, os.X_OK):
        if isdir:
            raise exception("Cannot traverse directory '%s':"
                            " lacks 'x' permission." % path)
        else:
            raise exception("File '%s' lacks execute ('x') permission." % path)
    return True


def deploy_configuration_file(filename, template_filename=None):
    """
    Ensure that configuration file `filename` exists; possibly
    copying it from the specified `template_filename`.

    Return `True` if a file with the specified name exists in the
    configuration directory.  If not, try to copy the template file
    over and then return `False`; in case the copy operations fails,
    a `NoConfigurationFile` exception is raised.

    The `template_filename` is always resolved relative to GC3Libs'
    'package resource' directory (i.e., the ``etc/`` directory in the
    sources.  If `template_filename` is `None`, then it is assumed to
    be the base name of `filename`.
    """
    if template_filename is None:
        template_filename = os.path.basename(filename)
    if not os.path.isabs(filename):
        filename = os.path.abspath(filename)
    if os.path.exists(filename):
        return True
    else:
        try:
            from pkg_resources import (Requirement, resource_filename,
                                       DistributionNotFound)
            import shutil
        except ImportError as err:
            raise gc3libs.exceptions.FatalError(
                "Cannot import required Python modules: %s."
                " Please check GC3Pie installation instructions at"
                " http://gc3pie.googlecode.com/svn/trunk/gc3pie/docs/html/install.html" %  # noqa
                str(err))
        try:
            # copy sample config file
            if not os.path.exists(dirname(filename)):
                os.makedirs(dirname(filename))
            sample_config = resource_filename(
                Requirement.parse("gc3pie"),
                "gc3libs/etc/" +
                template_filename)
            shutil.copyfile(sample_config, filename)
            return False
        except IOError as err:
            gc3libs.log.critical("Failed copying configuration file: %s", err)
            raise gc3libs.exceptions.NoConfigurationFile(
                "No configuration file '%s' was found, and an attempt to"
                " create it failed. Aborting." % filename)
        except DistributionNotFound as ex:
            raise AssertionError(
                "Cannot access resources for Python package: %s."
                " Installation error?" %
                str(ex))


def dirname(pathname):
    """
    Same as `os.path.dirname` but return `.` in case of path names with no
    directory component.
    """
    # FIXME: figure out if this is a desirable outcome.  i.e. do we
    # want dirname to be empty, or do a pwd and find out what the
    # current dir is, or keep the "./".  I suppose this could make a
    # difference to some of the behavior of the scripts, such as
    # copying files around and such.
    return os.path.dirname(pathname) or '.'


class Enum(frozenset):
    """
    A generic enumeration class.  Inspired by: http://goo.gl/1AL5N0
    with some more syntactic sugar added.

    An `Enum` class must be instanciated with a list of strings, that
    make the enumeration "label"::

      >>> Animal = Enum('CAT', 'DOG')

    Each label is available as an instance attribute, evaluating to
    itself::

      >>> Animal.DOG
      'DOG'

      >>> Animal.CAT == 'CAT'
      True

    As a consequence, you can test for presence of an enumeration
    label by string value::

      >>> 'DOG' in Animal
      True

    Finally, enumeration labels can also be iterated upon::

      >>> for a in sorted(Animal): print(a)
      CAT
      DOG
    """
    def __new__(cls, *args):
        return frozenset.__new__(cls, args)

    def __getattr__(self, name):
        if name in self:
            return name
        else:
            raise AttributeError("No '%s' in enumeration '%s'"
                                 % (name, self.__class__.__name__))

    def __setattr__(self, name, value):
        raise SyntaxError("Cannot assign enumeration values.")

    def __delattr__(self, name):
        raise SyntaxError("Cannot delete enumeration values.")


class ExponentialBackoff(object):
    """Generate waiting times with the `exponential backoff`_ algorithm.

    Returned times are in seconds (or fractions thereof); they are
    integral multiples of the basic time slot, which is set with the
    `slot_duration` constructor parameter.

    After `max_retries` have been attempted, any call to this iterator
    will raise a `StopIteration` exception.

    The `ExponentialBackoff` class implements the iterator protocol,
    so you can just retrieve waiting times with the `.next()` method,
    or by looping over it::

      >>> lapses = list(ExponentialBackoff(max_retries=7))
      >>> len(lapses)
      8

    .. _`exponential backoff`: http://goo.gl/PxVICA

    """

    def __init__(self, slot_duration=0.05, max_retries=5):
        self.attempt = -1
        self.slot_duration = slot_duration
        self.max_retries = max_retries

    def __iter__(self):
        return self

    def __next__(self):
        """Return next waiting time."""
        self.attempt += 1
        if self.attempt > self.max_retries:
            raise StopIteration
        return self.slot_duration * random.randint(0, 2 ** self.attempt - 1)

    def wait(self):
        """Wait for another while."""
        time.sleep(next(self))


def first(seq):
    """
    Return the first element of sequence or iterator `seq`.
    Raise `TypeError` if the argument does not implement
    either of the two interfaces.

    Examples::

      >>> s = [0, 1, 2]
      >>> first(s)
      0

      >>> s = {'a':1, 'b':2, 'c':3}
      >>> first(sorted(s.keys())) == 'a'
      True
    """
    try:  # try iterator interface
        return next(seq)
    except TypeError:
        pass
    try:  # seq is no iterator, try indexed lookup
        return seq[0]
    except IndexError:
        pass
    raise TypeError(
        "Argument to `first()` method needs to be iterator or sequence.")


def fgrep(literal, filename):
    """
    Iterate over all lines in a file that contain the `literal` string.
    """
    with open(filename, 'r') as stream:
        for line in stream:
            if literal in line:
                yield line


def from_template(template, **extra_args):
    """
    Return the contents of `template`, substituting all occurrences
    of Python formatting directives '%(key)s' with the corresponding values
    taken from dictionary `extra_args`.

    If `template` is an object providing a `read()` method, that is
    used to gather the template contents; else, if a file named
    `template` exists, the template contents are read from it;
    otherwise, `template` is treated like a string providing the
    template contents itself.
    """
    if hasattr(template, 'read') and callable(template.read):
        template_contents = template.read()
    elif os.path.exists(template):
        template_file = file(template, 'r')
        template_contents = template_file.read()
        template_file.close()
    else:
        # treat `template` as a string
        template_contents = template
    # substitute `extra_args` into `t` and return it
    return (template_contents % extra_args)


def from_encoded_bytes(chars, encodings):
    """
    Convert *chars* to Python unicode string, trying different encodings.

    Try converting byte string *chars* to a Python text string (type
    ``unicode`` on Py2, type ``str`` on Py3), trying each of the
    encodings specified until one of them succeeds.

    If none of the encodings work, raise ``UnicodeDecodeError``.

    :param bytes chars:
      Byte string to convert to text
    :param list encodings:
      List of encodings to try, in sequence. (e.g., ``['utf-8', 'latin-1']``
    :raises UnicodeDecodeError:
      When none of the encodings can successfully convert the given byte string.
    """
    for encoding in encodings:
        try:
            return decode(chars, encoding)
        except UnicodeDecodeError:
            pass
    raise UnicodeDecodeError(
        "Cannot decode byte string {0!r}; tried encodings {1!r}."
        .format(chars, encodings))


def from_filesystem_bytes(chars):
    """
    Convert *chars* to Python unicode string, trying different encodings.

    This function should be used to make a Python text string (type
    ``unicode`` on Python 2, type ``str`` on Pythoin 3) out of a byte
    string of characters that result from a filesystem lookup
    operation.  Conversion to a text string is attempted using the
    following encodings, in order:

    #. the encoding determined by the current locale
       (as determined by Python's ``locale.getlocale()``);
    #. the "user's preferred encoding", as determined
       by Python's ``locale.getpreferredencoding()``;
    #. UTF-8 encoding;
    #. direct map of byte values 0x0 through 0xff to the corresponding
       Unicode code points.

    The latter conversion will *not* be a valid text conversion (i.e.,
    it will not preserve any text representation of the string),
    unless the ISO-8859-1 (aka "latin-1") encoding is used.

    However, the issue we need to solve here is the mismatch of
    Python's use of text strings to represent path names with UNIX' C
    library use of byte strings to represent the same.  Since the UNIX
    kernel has no notion of character encoding in path names (as
    opposed to Windows, which uses UNICODE since Win95), there is
    actually no guarantee that a given file name can be decoded.  This
    is particularly evident on multi-user Linux/UNIX systems where you
    may be using e.g. the ``en_US.UTF-8`` locale but your Russian
    colleague may be using KOI8-R...

    :param bytes chars:
      Byte string to convert to text
    :raises UnicodeDecodeError:
      When none of the encodings can successfully convert the given byte string.
    """
    encodings = [ 'utf-8', 'latin-1']
    preferred = locale.getpreferredencoding()
    if preferred:
        encodings.insert(0, preferred)
    default = locale.getlocale()[1]
    if default:
        encodings.insert(0, default)
    return from_encoded_bytes(chars, encodings)


def from_terminal_bytes(chars):
    """
    Convert *chars* to Python unicode string, using current locale encoding.

    This function should be used to make a Python text string (type
    ``unicode`` on Python 2, type ``str`` on Pythoin 3) out of a byte
    string of characters that were inputed by users in a terminal
    application (e.g., returned by ``input()`` or typed as
    command-line arguments).

    If the current locale encoding cannot be determined, then the byte
    string is assumed to be an ASCII-only string.

    :param bytes chars:
      Byte string to convert to text
    :raises UnicodeDecodeError:
      When none of the encodings can successfully convert the given byte string.
    """
    return decode(chars, locale.getlocale()[1] or 'ascii')



def get_available_physical_memory():
    """
    Return size of available memory (as a `gc3libs.quantity.Memory` object).
    The figure only refers to RAM, i.e., *physical* memory as opposed
    to *virtual* memory (swap).

    Should work on any POSIX system that supports the
    ``_SC_AVPHYS_PAGES`` variable in the ``sysconf()`` C library call.

    :raise NotImplementedError:
      If syscalls to determine amount of available physical memory
      are not implemented on this system.
    """
    try:
        pagesize = os_sysconf('SC_PAGE_SIZE')
        avail_pages = os_sysconf('SC_AVPHYS_PAGES')
        return Memory(avail_pages * pagesize, unit=Memory.B)
    except ValueError:
        raise NotImplementedError(
            "Cannot determine amount of available physical memory.")


def get_linux_memcg_limit():
    """
    Return memory limit in this process' Linux memory cgroup.
    Return value is a `gc3libs.quantity.Memory` object,
    or ``None`` if no limit can be detected.

    As the Linux "memory cgroup" mechanism implements different
    limits, and not all of them might have been set/enforced, we we
    read possible limits in supposedly ascending order ("soft" limits
    *should* be lower than "hard" limits) and return first one that
    exists.

    See also: `<https://www.kernel.org/doc/Documentation/cgroup-v1/memory.txt>`_
    """
    memcg_path = None
    try:
        for line in fgrep(':memory:', '/proc/self/cgroup'):
            _, _, memcg_path = line.rstrip().split(':')
            break
    except EnvironmentError:
        # older kernel, no file `/proc/self/cgroup`
        pass
    if memcg_path is None:
        # no memory cgroup?
        return None
    # XXX: hard-coded cgroupfs mountpoint path
    memcg_dir = os.path.join('/sys/fs/cgroup/memory', memcg_path.lstrip('/'))
    limit = None
    # read possible limits in ascending order ("soft" limits *should*
    # be lower than "hard" limits) and return first one that exists
    for limit_file in [
            'memory.soft_limit_in_bytes',
            'memory.limit_in_bytes',
            'memory.memsw.limit_in_bytes',
    ]:
        path = os.path.join(memcg_dir, limit_file)
        try:
            with open(path, 'r') as limit_file:
                limit = int(limit_file.read())
                break
        except (EnvironmentError, ValueError):
            pass
    if limit:
        return Memory(limit, unit=Memory.B)
    else:
        return None


def get_max_real_memory():
    """
    Return maximum size of available *real* memory.
    Return value is a `gc3libs.quantity.Memory` object,
    or ``None`` if no limit can be detected.

    Various sources are polled for a limit, and the minimum is returned:

    - *available physical memory*, as reported by sysconf(3);
    - current resource limits (also known as ``ulimit`` in shell programming)
      ``RLIMIT_DATA`` and ``RLIMIT_AS``;
    - current Linux memory cgroup limits.
    """
    limits = []
    # Python's `posix` module does not expose an interface to the libc
    # `getrlimit()` call, so we resort to parsing `/proc/self/limits`
    # if available
    try:
        ulimit, _ = parse_linux_proc_limits()
        rlimit_as = ulimit['max_address_space']
        if rlimit_as:
            limits.append(rlimit_as)
        rlimit_data = ulimit['max_data_size']
        if rlimit_data:
            limits.append(rlimit_data)
    except (IOError, ValueError):
        pass
    # try using sysconf()
    try:
        avail_ram = get_available_physical_memory()
        limits.append(avail_ram)
    except NotImplementedError:
        pass
    # try Linux memory cgroup
    memcg_limit = get_linux_memcg_limit()
    if memcg_limit:
        limits.append(memcg_limit)
    # whichever is lower
    if limits:
        return min(limits)
    else:
        return None


def get_num_processors():
    """
    Return number of online processor cores.
    """
    # try different strategies and use first one that succeeeds
    try:
        return os.cpu_count()  # Py3 only
    except AttributeError:
        pass
    try:
        import multiprocessing
        return multiprocessing.cpu_count()
    except ImportError:  # no multiprocessing?
        pass
    except NotImplementedError:
        # multiprocessing cannot determine CPU count
        pass
    return None


def get_scheduler_and_lock_factory(lib):
    """
    Return factories for creating a period task scheduler and locks.

    The scheduler will be a scheduler class from the APScheduler_
    framework (which see for the API), and the lock factory is an
    appropriate locking object for synchronizing independently running
    tasks. Example::

        sched_factory, lock_factory = _get_scheduler_and_lock_factory('threading')
        sched = sched_factory()
        sched.add_job(task1, 'interval', seconds=5)
        sched.add_job(task2, 'interval', seconds=30)

        shared_data_lock = lock_factory()

        def task1():
          # ...
          with shared_data_lock:
            # modify shared data

    Argument `lib` is one of: ``threading``, ``gevent``, ``tornado``,
    ``asyncio`` (Python 3.5+ "async" system), ``twisted``, ``qt``;
    each of them selects a scheduler and lock objects compatible with
    the named framework for concurrent processing.

    .. _APScheduler: https://apscheduler.readthedocs.org/en/latest/userguide.html
    """
    if lib == 'threading':
        from apscheduler.schedulers.background import BackgroundScheduler
        from threading import Lock
        return (BackgroundScheduler, Lock)
    elif lib == 'gevent':
        from apscheduler.schedulers.gevent import GeventScheduler
        from gevent.lock import Semaphore
        return (GeventScheduler, Semaphore)
    elif lib in ['asyncio', 'tornado', 'twisted', 'qt']:
        raise NotImplemented(
            "Support for {lib} is not yet available!"
            .format(lib=lib))
    else:
        raise ValueError(
            "Library '{lib}' is unknown to `{mod}._get_scheduler_and_lock_factory()`"
            .format(lib=lib, mod=__name__))


def getattr_nested(obj, name):
    """
    Like Python's `getattr`, but perform a recursive lookup if `name` contains
    any dots.
    """
    dots = name.count('.')
    if dots == 0:
        return getattr(obj, name)
    else:
        first, rest = name.split('.', 1)
        return getattr_nested(getattr(obj, first), rest)


def grep(pattern, filename):
    """
    Iterate over all lines in a file that match the `pattern` regular
    expression.
    """
    rx = re.compile(pattern)
    with open(filename, 'r') as file:
        for line in file:
            if rx.search(line):
                yield line


def ifelse(test, if_true, if_false):
    """
    Return `if_true` is argument `test` evaluates to `True`,
    return `if_false` otherwise.

    This is just a workaround for Python 2.4 lack of the
    conditional assignment operator::

      >>> a = 1
      >>> b = ifelse(a, "yes", "no"); print(b)
      yes
      >>> b = ifelse(not a, 'yay', 'nope'); print(b)
      nope

    """
    if test:
        return if_true
    else:
        return if_false


# original source: https://gist.github.com/jtriley/7270594
def insert_char_every_n_chars(string, char='\n', every=64):
    return char.join(
        string[i:i + every] for i in range(0, len(string), every))


def irange(start, stop, step=1):
    """
    Iterate over all values greater or equal than `start` and less than `stop`.
    (Or the reverse, if `step < 0`.)

    Example::

      >>> list(irange(1, 5))
      [1, 2, 3, 4]
      >>> list(irange(0, 8, 3))
      [0, 3, 6]
      >>> list(irange(8, 0, -2))
      [8, 6, 4, 2]

    Unlike the built-in `range` function, `irange` also accepts
    floating-point values::

      >>> list(irange(0.0, 1.0, 0.5))
      [0.0, 0.5]

    Also unlike the built-in `range`, *both* `start` and `stop` have
    to be specified::

      >>> try:
      ...   irange(42)
      ... except TypeError:
      ...   print("missing required argument!")
      missing required argument!

    Of course, a null `step` is not allowed::

      >>> try:
      ...   list(irange(1, 2, 0))
      ... except AssertionError as err:
      ...   assert 'Null step in irange.' in str(err)
    """
    assert float(step) != 0.0, "Null step in irange."
    value = start
    if step > 0:
        while value < stop:
            yield value
            value += step
    else:  # step < 0
        while value > stop:
            yield value
            value += step


def lock(path, timeout, create=True):
    """
    Lock the file at `path`.  Raise a `LockTimeout` error if the lock
    cannot be acquired within `timeout` seconds.

    Return a `lock` object that should be passed unchanged to the
    `gc3libs.utils.unlock` function.

    If no `path` points to a non-existent location, an empty file is
    created before attempting to lock (unless `create` is `False`).
    An attempt is made to remove the file in case an error happens.

    See also: `gc3libs.utils.unlock`:func:
    """
    # ``FileLock`` requires that the to-be-locked file exists; if it
    # does not, we create an empty one (and avoid overwriting any
    # content, in case another process is also writing to it).  There
    # is thus no race condition here, as we attempt to lock the file
    # anyway, and this will stop concurrent processes.
    if not os.path.exists(path) and create:
        open(path, "a").close()
        created = True
    else:
        created = False
    try:
        lck = lockfile.FileLock(path, threaded=False)
        lck.acquire(timeout=timeout)
    except Exception:
        if created:
            try:
                os.remove(path)
            except:
                pass
        raise
    return lck


class History(object):
    """
    A list of messages with timestamps and (optional) tags.

    The `append` method should be used to add a message to the `History`::

      >>> L = History()
      >>> L.append('first message')
      >>> L.append('second one')

    The `last` method returns the text of the last message appended, with its
    timestamp::

      >>> L.last().startswith('second one at')
      True

    Iterating over a `History` instance returns message texts in the
    temporal order they were added to the list, with their timestamp::

      >>> for msg in L: print(msg) # doctest: +ELLIPSIS
      first message ...

    """

    def __init__(self):
        self._messages = []

    def append(self, message, *tags):
        """
        Append a message to this `History`.

        The message is timestamped with the time at the moment of the
        call.

        The optional `tags` argument is a sequence of strings. Tags
        are recorded together with the message and may be used to
        filter log messages given a set of labels. *(This feature is
        not yet implemented.)*

        """
        self._messages.append((message, time.time(), tags))

    def last(self):
        """
        Return text of last message appended.
        If log is empty, return empty string.
        """
        if len(self._messages) == 0:
            return ''
        else:
            return self.format_message(self._messages[-1])

    def format_message(self, message):
        """Return a formatted message, appending to the message its timestamp
        in human readable format.

        """
        return "%s at %s" % (
            message[0], time.asctime(time.localtime(message[1])))

    # shortcut for append
    def __call__(self, message, *tags):
        """Shortcut for `History.append` (which see)."""
        self.append(message, *tags)

    def __iter__(self):
        """Iterate over messages in the temporal order they were added."""
        return iter([self.format_message(record) for record in self._messages])

    def __str__(self):
        """Return all messages texts in a single string, separated by newline
        characters."""
        return '- ' + '\n- '.join([self.format_message(record)
                               for record in self._messages]) + '\n'


def lookup(obj, name):
    """
    Return attribute or item with the given name in collection `obj`.

    :raise LookupError:
      If `obj` has no attribute nor item with the given name.

    This is meant for cases when different versions of an API may
    either return a dictionary (hence, key/``__getitem__``-based
    lookup) or an object/namespace (hence, ``.``/``getattr``-style
    lookup) and you want to handle them in a uniform way.

    The following examples demo it::

      >>> D = {'a':1, 'b':2}
      >>> lookup(D, 'a')
      1
      >>> lookup(D, 'c')
      Traceback (most recent call last):
        ...
      LookupError: Object ... has no attribute nor key named `c`

      >>> class X(object):
      ...   a = 1
      ...   b = 2
      >>> x = X()
      >>> lookup(x, 'a')
      1
      >>> lookup(x, 'c')
      Traceback (most recent call last):
        ...
      LookupError: Object ... has no attribute nor key named `c`
    """
    try:
        return getattr(obj, name)
    except AttributeError:
        pass
    try:
        return obj[name]
    except (KeyError, TypeError):
        pass
    raise LookupError(
        'Object {0} has no attribute nor key named `{1}`'
        .format(obj, name))


def mkdir(path, mode=0o777):
    """
    Like `os.makedirs`, but does not throw an exception if PATH
    already exists.
    """
    if not os.path.exists(path):
        os.makedirs(path, mode)


def mkdir_with_backup(path, mode=0o777):
    """
    Like `os.makedirs`, but if `path` already exists and is not empty,
    rename the existing one to a backup name (see the `backup` function).

    Unlike `os.makedirs`, no exception is thrown if the directory
    already exists and is empty, but the target directory permissions
    are not altered to reflect `mode`.
    """
    if os.path.isdir(path):
        if len(os.listdir(path)) > 0:
            # directory already exists and is non-empty; backup it and
            # make a new one
            backup(path)
            os.makedirs(path, mode)
        else:
            # keep existing empty directory
            pass
    else:
        os.makedirs(path, mode)


def movefile(src, dst, overwrite=False, changed_only=True, link=False):
    """
    Move a file from `src` to `dst`; return `True` if the move was
    actually made.

    The `overwrite` and `changed_only` optional arguments
    have the same effect as in `copyfile`:func: (which see).

    If `dst` is a directory, a file with the same basename as `src` is
    created (or overwritten) in the directory specified.

    Return ``True`` or ``False``, depending on whether the source file
    was actually moved to the destination.

    See also: `copyfile`:func:
    """
    if os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    if os.path.exists(dst) and not overwrite:
        return False
    if samefile(src, dst):
        return False
    if not os.path.exists(dst):
        dstdir = dirname(dst)
        if not os.path.exists(dstdir):
            os.makedirs(dstdir)
    else:
        # `dst` exists, check for changes
        if changed_only:
            sstat = os.stat(src)
            dstat = os.stat(dst)
            if (sstat.st_size ==
                    dstat.st_size and sstat.st_mtime <= dstat.st_mtime):
                # same size and destination more recent, do not move
                return False
    try:
        shutil.move(src, dst)
    except WindowsError:
        pass
    return True


def movetree(src, dst, overwrite=False, changed_only=True):
    """
    Recursively move an entire directory tree rooted at `src`.

    The `overwrite` and `changed_only` optional arguments
    have the same effect as in `copytree`:func: (which see).

    See also: `copytree`:func:.
    """
    assert os.path.isdir(src), \
        ("Source path `%s` does not name an existing directory" % src)
    errors = []
    if not os.path.exists(dst):
        os.makedirs(dst)
    for name in os.listdir(src):
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if os.path.isdir(srcname):
                errors.extend(
                    movetree(srcname, dstname, overwrite, changed_only))
            else:
                movefile(srcname, dstname)
        except (IOError, os.error) as why:
            errors.append((srcname, dstname, why))
    return errors


def move_recursively(src, dst, overwrite=False, changed_only=True):
    """
    Move `src` to `dst`, descending it recursively if necessary.

    The `overwrite` and `changed_only` optional arguments
    have the same effect as in `copytree`:func: (which see).
    """
    if os.path.isdir(src):
        movetree(src, dst, overwrite, changed_only)
    else:
        movefile(src, dst, overwrite, changed_only)


def occurs(pattern, filename, match=grep):
    """
    Return ``True`` if a line in `filename` matches `pattern`.

    The `match` argument selects how exactly `pattern` is searched for
    in the contents of `filename`:

    * when `match=grep` (default), then `pattern` is a regular
      expression that is searched for (unanchored) in every line;

    * when `match=fgrep`, then `pattern` is a string that is searched
      for literally in every line;

    * more in general, the `match` function should return an iterator
      over matches of `pattern` within the contents of `filename`: if
      at least one match is found, `occurs` will return ``True``.

    :param str pattern:  Pattern to search for
    :param str filename: Path name of the file to search into
    :param match: Function returning iterator over matches
    """
    try:
        # look for the first match -- if one is found, we're done
        next(match(pattern, filename))
        return True
    except StopIteration:
        return False


def parse_linux_proc_limits(data=None):
    """
    Return dictionary mapping limit name to corresponding value.
    In case the actual limit is 'unlimited', value is set to ``None``.
    """
    if data is None:
        with open('/proc/self/limits', 'r') as src:
            data = src.read()
    lines = data.split('\n')
    assert lines
    # Linux' `/proc/self/limits` shows data in 4 columns, but: (1)
    # value in the first column may contain spaces, and (2) the value
    # in the last column may be missing.  However, values in the same
    # column (incl. header) are left-aligned on the same screen
    # position (for human reading) so we need only note the starting
    # position for each data column.
    header = lines[0]
    name_col = header.index('Limit')
    soft_limit_col = header.index('Soft Limit')
    hard_limit_col = header.index('Hard Limit')
    unit_col = header.index('Units')
    # parse data line-by-line
    soft_limits = {}
    hard_limits = {}
    for line in lines[1:]:
        if line == '':
            continue
        name = line[name_col:soft_limit_col].rstrip()
        soft_limit = line[soft_limit_col:hard_limit_col].rstrip()
        if unit_col < len(line):
            hard_limit = line[hard_limit_col:unit_col].rstrip()
            unit = line[unit_col:].rstrip()
        else:
            # no value for `Unit`
            hard_limit = line[hard_limit_col:].rstrip()
            unit = None
        name = name.lower().replace(' ', '_')
        soft_limits[name] = _parse_ulimit(soft_limit, unit)
        hard_limits[name] = _parse_ulimit(hard_limit, unit)
    return soft_limits, hard_limits


def _parse_ulimit(limit, unit):
    """
    Helper function for `parse_linux_proc_limits`.
    """
    if limit == 'unlimited':
        return None
    else:
        limit = int(limit)
    # translate to `gc3libs.quantity` unit
    if unit == 'bytes':
        cls = Memory
        unit = Memory.B
    elif unit == 'seconds':
        cls = Duration
        unit = Duration.s
    elif unit == 'us':
        cls = Duration
        unit = Duration.us
    else:
        # "files", "signals", "processes", etc.
        cls = (lambda v, u: v)
        unit = None
    return cls(limit, unit)


def parse_range(spec):
    """
    Return minimum, maximum, and stepping value for a range.

    Argument `spec` must be a string of the form `LOW:HIGH:STEP`,
    where LOW, HIGH and STEP are (integer or floating-point) numbers.
    Example::

      >>> parse_range('1:10:2')
      (1, 10, 2)

      >>> parse_range('1.0:3.5:0.5')
      (1.0, 3.5, 0.5)

    Note that, as soon as *any* one of LOW, HIGH, STEP is not an
    integer, *all* of them are parsed as Python floats::

      >>> parse_range('1:3:0.5')
      (1.0, 3.0, 0.5)

      >>> parse_range('1.0:3:1')
      (1.0, 3.0, 1.0)

      >>> parse_range('1:3.0:1')
      (1.0, 3.0, 1.0)

    The final part `:STEP` can be omitted if the step is `1`::

      >>> parse_range('2:5')
      (2, 5, 1)

      >>> parse_range('1.0:3.0')
      (1.0, 3.0, 1.0)

    Finally, note that `parse_range` does not perform any kind of
    check on the validity of the resulting range; so it is possible to
    parse a string into an empty range or range specification with
    stepping 0::

      >>> parse_range('1:-5:10')
      (1, -5, 10)

      >>> parse_range('1:2:0')
      (1, 2, 0)

    As a special case to simplify user interfaces, a single number is
    accepted as a *degenerate* range: it will be parsed as a range
    whose content is just the given number::

      >>> parse_range('42')
      (42, 43, 1)
    """
    colons = spec.count(':')
    if colons == 2:
        low, high, step = spec.split(':')
    elif colons == 1:
        low, high = spec.split(':')
        step = '1'  # parsed to int or float later on
    elif colons == 0:
        low = spec
        if '.' in low:
            high = str(float(low) + 1)
        else:
            high = str(int(low) + 1)
        step = '1'  # parsed to int or float later on
    else:
        raise ValueError(
            "Argument `spec` must have the form 'LOW:HIGH:STEP',"
            " where LOW, HIGH and STEP are (integer or"
            " floating-point) numbers.")
    # are low, high, step to floats or ints?
    if ('.' in low) or ('.' in high) or ('.' in step):
        low = float(low)
        high = float(high)
        step = float(step)
    else:
        low = int(low)
        high = int(high)
        step = int(step)
    return low, high, step


def prettyprint(
        D,
        indent=0,
        width=0,
        maxdepth=None,
        step=4,
        only_keys=None,
        output=sys.stdout,
        _key_prefix='',
        _exclude=None):
    """
    Print dictionary instance `D` in a YAML-like format.
    Each output line consists of:

      * `indent` spaces,
      * the key name,
      * a colon character ``:``,
      * the associated value.

    If the total line length exceeds `width`, the value is printed
    on the next line, indented by further `step` spaces; a value of 0 for
    `width` disables this line wrapping.

    Optional argument `only_keys` can be a callable that must return
    `True` when called with keys that should be printed, or a list of
    key names to print.

    Dictionary instances appearing as values are processed recursively
    (up to `maxdepth` nesting).  Each nested instance is printed
    indented `step` spaces from the enclosing dictionary.
    """
    # be sure we do not try to recursively dump `D`
    if _exclude is None:
        _exclude = set()
    _exclude.add(id(D))
    for k, v in sorted(D.items()):
        leading_spaces = indent * ' '
        full_name = "%s%s" % (_key_prefix, k)
        if only_keys is not None:
            try:
                # is `only_keys` a filter function?
                if not only_keys(str(full_name)):
                    continue
            except TypeError:
                # no, then it must be a list of key names, check for
                # keys having the same number of dots as in the prefix
                level = _key_prefix.count('.')
                found = False
                for name in only_keys:
                    # take only the initial segment, up to a "level" dots
                    dots = min(name.count('.'), level) + 1
                    prefix = '.'.join(name.split('.')[:dots])
                    if str(full_name) == prefix:
                        found = True
                        break
                if not found:
                    continue
        # ignore excluded items
        if id(v) in _exclude:
            continue
        # To make a 'key' valid in YAML it must not start with one of the following chars
        sk = str(k)
        sk = sk if sk[0] not in  u'\0 \t\r\n\x85\u2028\u2029-?:,[]{}#&*!|>\'\"%@`' else  "'%s'" % sk
        first = ''.join([leading_spaces, sk, ': '])
        if isinstance(
                v, (dict, Mapping, OrderedDict)):
            if maxdepth is None or maxdepth > 0:
                if maxdepth is None:
                    depth = None
                else:
                    depth = maxdepth - 1
                sstream = StringIO.StringIO()
                prettyprint(v, indent + step, width, depth, step,
                            only_keys, sstream, full_name + '.', _exclude)
                second = sstream.getvalue()
                sstream.close()
            elif maxdepth == 0:
                second = "..."
        elif isinstance(v, (list, tuple)):
            second = ', '.join([str(item) for item in v])
        else:
            second = str(v)
        # wrap overlong lines, and always wrap if the second part is multi-line
        if (width > 0 and len(first) + len(second)
                > width) or ('\n' in second):
            first += '\n'
        # indent a multi-line block by indent+step spaces
        if '\n' in second:
            lines = second.splitlines()
            # keep indentation relative to first line
            dedent = 0
            line0 = lines[0].expandtabs(step)
            while line0[dedent].isspace():
                dedent += 1
            # rebuild `second`, indenting each line by (indent+step) spaces
            second = ''
            for line in lines:
                second = ''.join([
                    second,
                    ' ' * (indent + step),
                    line.rstrip().expandtabs(step)[dedent:],
                    '\n'
                ])
        # there can be multiple trailing '\n's, which we remove here
        second = second.rstrip()
        # finally print line(s)
        output.write('%s%s\n' % (first, second))


def progressive_number(qty=None, id_filename=None):
    """
    Return a positive integer, whose value is guaranteed to
    be monotonically increasing across different invocations
    of this function, and also across separate instances of the
    calling program.

    This is accomplished by using a system-wide file which holds the
    "next available" ID.  The location of this file can be set using
    the ``GC3PIE_ID_FILE`` environment variable, or programmatically
    using the `id_filename` argument.  By default, the "next ID" file
    is located at ``~/.gc3/next_id.txt``:file:

    Example::

      >>> # create "next ID" file in a temporary location
      >>> import tempfile, os
      >>> (fd, tmp) = tempfile.mkstemp()

      >>> n = progressive_number(id_filename=tmp)
      >>> m = progressive_number(id_filename=tmp)
      >>> m > n
      True

    If you specify a positive integer as argument, then a list of
    monotonically increasing numbers is returned.  For example::

      >>> ls = progressive_number(5, id_filename=tmp)
      >>> len(ls)
      5

    (clean up test environment)
      >>> os.remove(tmp)

    In other words, `progressive_number(N)` is equivalent to::

      nums = [ progressive_number() for n in range(N) ]

    only more efficient, because it has to obtain and release the lock
    only once.

    After every invocation of this function, the last returned number
    is stored into the file passed as argument `id_filename`.  If the
    file does not exist, an attempt to create it is made before
    allocating an id; the method can raise an `IOError` or `OSError`
    if `id_filename` cannot be opened for writing.

    *Note:* as file-level locking is used to serialize access to the
    counter file, this function may block (default timeout: 30
    seconds) while trying to acquire the lock, or raise a
    `LockTimeout` exception if this fails.

    :raise: LockTimeout, IOError, OSError

    :returns: A positive integer number, monotonically increasing with
            every call.  A list of such numbers if argument `qty` is a
            positive integer.
    """
    assert qty is None or qty > 0, \
        "Argument `qty` must be a positive integer"
    if id_filename is None:
        id_filename = os.environ.get('GC3PIE_ID_FILE',
                                     os.path.expanduser('~/.gc3/next_id.txt'))
    # ensure directory exists, otherwise the error message is obscure;
    # see Issue 486 for details
    id_dirname = dirname(id_filename)
    if not os.path.exists(id_dirname):
        os.makedirs(id_dirname)
    # XXX: can raise 'LockTimeout'
    lck = lock(id_filename, timeout=30, create=True)
    id_file = open(id_filename, 'r+')
    id = int(id_file.read(8) or "0", 16)
    id_file.seek(0)
    if qty is None:
        id_file.write(
            "%08x -- DO NOT REMOVE OR ALTER THIS FILE: it is used internally"
            " by the gc3libs\n" % (id + 1))
    else:
        id_file.write(
            "%08x -- DO NOT REMOVE OR ALTER THIS FILE: it is used internally"
            " by the gc3libs\n" % (id + qty))
    id_file.close()
    unlock(lck)
    if qty is None:
        return id + 1
    else:
        return [(id + n) for n in range(1, qty + 1)]


def read_contents(path):
    """
    Return the whole contents of the file at `path` as a single string.

    Example::

      >>> read_contents('/dev/null')
      ''

      >>> import tempfile
      >>> (fd, tmpfile) = tempfile.mkstemp()
      >>> w = open(tmpfile, 'w')
      >>> w.write('hey') and None  # make doctest compatible with Py2 and Py3
      >>> w.close()
      >>> read_contents(tmpfile)
      'hey'

    (If you run this test, remember to do cleanup afterwards)
    ::

      >>> os.remove(tmpfile)

    """
    with open(path, 'r') as stream:
        return stream.read()


def remove(path):
    """
    Remove a file, but raise no exception if it does not exist.
    """
    try:
        if os.path.exists(path):
            os.remove(path)
    except Exception:
        pass


def safe_repr(obj):
    """
    Return a string describing Python object `obj`.

    Avoids calling any Python magic methods, so should be safe to use
    as a 'last resort' in implementation of `__str__` and `__repr__`.
    """
    return ("%s@%x" % (obj.__class__.__name__, id(obj)))


def same_docstring_as(referenced_fn):
    """
    Function decorator: sets the docstring of the following function
    to the one of `referenced_fn`.

    Intended usage is for setting docstrings on methods redefined in
    derived classes, so that they inherit the docstring from the
    corresponding abstract method in the base class.
    """
    def decorate(f):
        f.__doc__ = referenced_fn.__doc__
        return f
    return decorate


def samefile(path1, path2):
    """
    Like `os.path.samefile` but return `False` if either one of the
    paths does not exist.
    """
    try:
        return os.path.samefile(path1, path2)
    except OSError as err:
        if err.errno == 2:  # ENOENT
            return False
        raise


def sh_quote_safe(arg):
    """
    Escape a string for safely passing as argument to a shell command.

    Return a single-quoted string that expands to the exact literal
    contents of `text` when used as an argument to a shell command.
    Examples (note that backslashes are doubled because of Python's
    string read syntax)::

      >>> print(sh_quote_safe("arg"))
      'arg'
      >>> print(sh_quote_safe("'arg'"))
      ''\\''arg'\\'''

    """
    return ("'" + str(arg).replace("'", r"'\''") + "'")


def sh_quote_safe_cmdline(args):
    """
    Single-quote a list of strings for passing to the shell as a
    command.  Return string comprised of the quoted arguments,
    concatenated and separated by spaces.

    Examples::

      >>> print(sh_quote_safe_cmdline(['sh', '-c', 'echo c(1,2,3)']))
      'sh' '-c' 'echo c(1,2,3)'
    """
    return ' '.join((sh_quote_safe(arg) for arg in args))


_DQUOTE_RE = re.compile(r'(\\*)"')
"""Regular expression for escaping double quotes in strings."""


def sh_quote_unsafe(arg):
    """
    Double-quote a string for passing as argument to a shell command.

    Return a double-quoted string that expands to the contents of
    `text` but still allows variable expansion and ``\\``-escapes
    processing by the UNIX shell.  Examples (note that backslashes are
    doubled because of Python's string read syntax)::

      >>> print(sh_quote_unsafe("arg"))
      "arg"
      >>> print(sh_quote_unsafe('"arg"'))
      "\\"arg\\""
      >>> print(sh_quote_unsafe(r'"\\"arg\\""'))
      "\\"\\\\\\"arg\\\\\\"\\""

    """
    return ('"' + _DQUOTE_RE.sub(r'\1\1\"', str(arg)) + '"' )


def sh_quote_unsafe_cmdline(args):
    """
    Double-quote a list of strings for passing to the shell as a
    command.  Return string comprised of the quoted arguments,
    concatenated and separated by spaces.

    Examples::

      >>> print(sh_quote_unsafe_cmdline(['sh', '-c', 'echo $HOME']))
      "sh" "-c" "echo $HOME"
    """
    return ' '.join((sh_quote_unsafe(arg) for arg in args))


@total_ordering
class PlusInfinity(object):

    """
    An object that is greater-than any other object.

        >>> x = PlusInfinity()

        >>> x > 1
        True
        >>> 1 < x
        True
        >>> 1245632479102509834570124871023487235987634518745 < x
        True

        >>> x > sys.maxsize
        True
        >>> x < sys.maxsize
        False
        >>> sys.maxsize < x
        True

    `PlusInfinity` objects are actually larger than *any* given Python
    object::

        >>> x > 'azz'
        True
        >>> x > object()
        True

    Relational operators try to return the correct value when
    comparing `PlusInfinity` to other instances of `PlusInfinity`::

        >>> y = PlusInfinity()
        >>> x < y
        False
        >>> x <= y
        True
        >>> x == y
        True
        >>> x >= y
        True
        >>> x > y
        False

    Finally, addition and subtraction of a finite number from
    `PlusInfinity` always results in `PlusInfinity`::

        >>> y = x - 1
        >>> x == y
        True
        >>> y = x + 1
        >>> x == y
        True

    Note that this used to be a singleton with special handling of 'is'.
    This was removed because it was apparently unused, and a lot of extra code.

    """
    def __gt__(self, other):
        return type(self) != type(other)

    def __eq__(self, other):
        return type(self) == type(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):
        return self

    def __sub__(self, other):
        if other == self:
            raise RuntimeError(
                "Undefined result subtracting +infinity from +infinity.")
        return self


# In Python 2.7 still, `DictMixin` is an old-style class; thus, we need
# to make `Struct` inherit from `object` otherwise we loose properties
# when setting/pickling/unpickling and *very importantly* the ability to
# use `@property` ...
class Struct(MutableMapping):
    """
    A `dict`-like object, whose keys can be accessed with the usual
    '[...]' lookup syntax, or with the '.' get attribute syntax.

    Examples::

      >>> a = Struct()
      >>> a['x'] = 1
      >>> a.x
      1
      >>> a.y = 2
      >>> a['y']
      2

    Values can also be initially set by specifying them as keyword
    arguments to the constructor::

      >>> a = Struct(z=3)
      >>> a['z']
      3
      >>> a.z
      3

    Like `dict` instances, `Struct`s have a `copy` method to get a
    shallow copy of the instance:

      >>> b = a.copy()
      >>> b.z
      3

    """

    def __init__(self, initializer=None, **extra_args):
        if initializer is not None:
            try:
                # initializer is `dict`-like?
                for name, value in initializer.items():
                    self[name] = value
            except AttributeError:
                # initializer is a sequence of (name,value) pairs?
                for name, value in initializer:
                    self[name] = value
        for name, value in extra_args.items():
            self[name] = value

    def copy(self):
        """Return a (shallow) copy of this `Struct` instance."""
        return Struct(self)

    # the `DictMixin` class defines all std `dict` methods, provided
    # that `__getitem__`, `__setitem__` and `keys` are defined.
    def __delitem__(self, name):
        del self.__dict__[name]

    def __getitem__(self, name):
        return self.__dict__[name]

    def __setitem__(self, name, val):
        self.__dict__[name] = val

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def keys(self):
        return list(self.__dict__.keys())


def string_to_boolean(word):
    """
    Convert `word` to a Python boolean value and return it.
    The strings `true`, `yes`, `on`, `1` (with any
    capitalization and any amount of leading and trailing
    spaces) are recognized as meaning Python `True`::

      >>> string_to_boolean('yes')
      True
      >>> string_to_boolean('Yes')
      True
      >>> string_to_boolean('YES')
      True
      >>> string_to_boolean(' 1 ')
      True
      >>> string_to_boolean('True')
      True
      >>> string_to_boolean('on')
      True

    Any other word is considered as boolean `False`::

      >>> string_to_boolean('no')
      False
      >>> string_to_boolean('No')
      False
      >>> string_to_boolean('Nay!')
      False
      >>> string_to_boolean('woo-hoo')
      False

    This includes also the empty string and whitespace-only::

      >>> string_to_boolean('')
      False
      >>> string_to_boolean('  ')
      False

    """
    return word.strip().lower() in ['true', 'yes', 'on', '1']


@contextlib.contextmanager
def tempdir(**kwargs):
    """
    A context manager for creating and then deleting a temporary directory.

    All arguments are passed unchanged to the `tempfile.mkdtemp`
    standand library function.

    (Original source and credits: http://stackoverflow.com/a/10965572/459543)
    """
    tmpdir = tempfile.mkdtemp(**kwargs)
    try:
        yield tmpdir
    finally:
        if os.path.isdir(tmpdir):
            shutil.rmtree(tmpdir, ignore_errors=True)


def to_bytes(s):
    """
    Convert string `s` to an integer number of bytes.  Suffixes like
    'KB', 'MB', 'GB' (up to 'YB'), with or without the trailing 'B',
    are allowed and properly accounted for.  Case is ignored in
    suffixes.

    Examples::

      >>> to_bytes('12')
      12
      >>> to_bytes('12B')
      12
      >>> to_bytes('12KB')
      12000
      >>> to_bytes('1G')
      1000000000

    Binary units 'KiB', 'MiB' etc. are also accepted:

      >>> to_bytes('1KiB')
      1024
      >>> to_bytes('1MiB')
      1048576

    """
    # get the multiplier
    k = 1024 if 'i' in s.lower() else 1000
    # strip bad stuff away now
    s = s.lower().strip(' ib')
    # if it is just a number, return as int
    if s.isdigit():
        return int(s)
    num, unit = s[:-1], s[-1]
    return int(num) * (k ** BYTE_SUFFIX_TO_POW[unit])


def to_str(arg, origin='ascii'):
    """
    Convert *arg* to a Python text string.

    If *arg* is already a text string (i.e., a ``unicode`` object in
    Python 2, and a ``str`` object in Python 3), then return it
    unchanged.  As an exception, if *arg* is ``None``, return
    ``None`` (unchanged).

    Second argument *origin* determines the handling of *arg* when
    *arg* is a byte-string:

    - if *origin* is ``'filesystem'`` then *arg* is converted to a
      text string using `from_filesystem_bytes`:func: (which see);

    - if *origin* is ``'terminal'`` then conversion of *arg* is
      attempted using `from_terminal_bytes`:func: (which see);

    - otherwise, *origin* is interpreted as an encoding name, and byte
      string *arg* is decoded using that encoding's rules.

    If *arg* is neither a text string nor a byte string, then
    conversion to string is attempted using Python's built-in
    ``str()`` function.
    """
    if arg is None:
        return None
    elif isinstance(arg, text_str):
        return arg  # unmodified
    elif isinstance(arg, bytes):
        if origin == 'filesystem':
            return from_filesystem_bytes(arg)
        elif origin == 'terminal':
            return from_terminal_bytes(arg)
        else:
            # interpret `origin` as encoding
            return decode(arg, origin)
    else:
        return text_str(arg)


def touch(path):
    """
    Ensure a regular file exists at `path`.

    If the file already exists, its access and modification time are
    updated.

    (This is a very limited and stripped down version of the ``touch``
    POSIX utility.)
    """
    open(path, 'a').close()


def unlock(lock):
    """
    Release a previously-acquired lock.

    Argument `lock` should be the return value of a previous
    `gc3libs.utils.lock` call.

    See also: `gc3libs.utils.lock`:func:
    """
    lock.release()


def update_parameter_in_file(path, var_in, new_val, regex_in):
    '''
    Updates a parameter value in a parameter file using predefined regular
    expressions in `_loop_regexps`.

    :param path:    Full path to the parameter file.
    :param var_in:  The variable to modify.
    :param new_val: The updated parameter value.
    :param regex:   Name of the regular expression that describes the format
                    of the parameter file.
    '''
    _loop_regexps = {
        'bar-separated': (r'([a-z]+[\s\|]+)'
                          r'(\w+)'  # variable name
                          r'(\s*[\|]+\s*)'  # bars and spaces
                          r'([\w\s\.,;\[\]\-]+)'  # value
                          r'(\s*)'),
        'space-separated': (r'(\s*)'
                            r'(\w+)'  # variable name
                            r'(\s+)'  # spaces (filler)
                            r'([\w\s\.,;\[\]\-]+)'  # values
                            r'(\s*)'),  # spaces (filler)
    }
    isfound = False
    if regex_in in _loop_regexps.keys():
        regex_in = _loop_regexps[regex_in]
    para_file_in = open(path, 'r')
    para_file_out = open(path + '.tmp', 'w')
    for line in para_file_in:
        if not line.rstrip():
            continue
        (a, var, b, old_val, c) = re.match(regex_in, line.rstrip()).groups()
        gc3libs.log.debug(
            "Read variable '%s' with value '%s' ...", var, old_val)
        if var == var_in:
            isfound = True
            upd_val = new_val
        else:
            upd_val = old_val
        para_file_out.write(a + var + b + upd_val + c + '\n')
    para_file_out.close()
    para_file_in.close()
    # move new modified content over the old
    os.rename(path + '.tmp', path)
    if not isfound:
        gc3libs.log.critical(
            'update_parameter_in_file could not find parameter'
            ' in specified file')


def write_contents(path, data):
    """
    Overwrite the contents of the file at `path` with the given data.
    If the file does not exist, it is created.

    Example::

      >>> import tempfile
      >>> (fd, tmpfile) = tempfile.mkstemp()
      >>> write_contents(tmpfile, 'big data here') and None # discard return value on Py3
      >>> read_contents(tmpfile)
      'big data here'

    (If you run this test, remember to clean up afterwards)
    ::

      >>> os.remove(tmpfile)

    """
    mode = 'wb' if isinstance(data, bytes) else 'w'
    with open(path, mode) as stream:
        return stream.write(data)


class YieldAtNext(object):

    """
    Provide an alternate protocol for generators.

    Wrap a Python generator object, and buffer the return values from
    `send` and `throw` calls, returning `None` instead. Return the
    yielded value --or raise the `StopIteration` exception-- upon the
    subsequent call to the `next` method.
    """

    __slots__ = ['_generator', '_saved', '_stop_iteration']

    def __init__(self, generator):
        self._generator = generator
        self._saved = deque()
        self._stop_iteration = False

    def __iter__(self):
        return self

    def __next__(self):
        if self._stop_iteration:
            raise StopIteration
        elif self._saved:
            return self._saved.popleft()
        else:
            return next(self._generator)

    def send(self, value):
        try:
            self._saved.append(self._generator.send(value))
        except StopIteration:
            self._stop_iteration = True

    def throw(self, *excinfo):
        try:
            self._saved.append(self._generator.throw(*excinfo))
        except StopIteration:
            self._stop_iteration = True


##
# Main
##

if __name__ == '__main__':
    import doctest
    doctest.testmod(name='utils',
                    optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)



# test_engine.py
# -*- coding: utf-8 -*-
#
#  Copyright (C) 2015, 2018, 2019  University of Zurich. All rights reserved.
#
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or (at your
#  option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# stdlib imports
from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import range
from collections import defaultdict
from io import StringIO
import os
import shutil
import tempfile
import re

import pytest

# GC3Pie imports
from gc3libs import Run, Application, create_engine
import gc3libs.config
from gc3libs.core import Core, Engine, MatchMaker
from gc3libs.persistence.filesystem import FilesystemStore
from gc3libs.quantity import GB, hours

from gc3libs.testing.helpers import SimpleParallelTaskCollection, SimpleSequentialTaskCollection, SuccessfulApp, temporary_config, temporary_config_file, temporary_core, temporary_directory, temporary_engine


def test_engine_resources():
    """
    Check that configured resources can be accessed through the `Engine` object.
    """
    with temporary_engine() as engine:
        resources = engine.resources
        assert len(resources) == 1
        assert 'test' in resources
        test_rsc = resources['test']
        # these should match the resource definition in `gc3libs.testing.helpers.temporary_core`
        assert test_rsc.max_cores_per_job == 123
        assert test_rsc.max_memory_per_core == 999*GB
        assert test_rsc.max_walltime == 7*hours


def test_engine_progress(num_jobs=5, transition_graph=None, max_iter=100):
    with temporary_engine() as engine:

        # generate some no-op tasks
        tasks = []
        for n in range(num_jobs):
            name = 'app{nr}'.format(nr=n+1)
            app = SuccessfulApp(name)
            engine.add(app)
            tasks.append(app)

        # run them all
        current_iter = 0
        done = engine.counts()[Run.State.TERMINATED]
        while done < num_jobs and current_iter < max_iter:
            engine.progress()
            done = engine.counts()[Run.State.TERMINATED]
            current_iter += 1

        # check state
        for task in tasks:
            assert task.execution.state == 'TERMINATED'


def test_engine_forget_terminated(num_jobs=3, transition_graph=None, max_iter=100):
    with temporary_engine() as engine:
        engine.forget_terminated = True

        # generate some no-op tasks
        tasks = []
        for n in range(num_jobs):
            name = 'app{nr}'.format(nr=n+1)
            app = SuccessfulApp(name)
            engine.add(app)
            tasks.append(app)

        # run them all
        current_iter = 0
        done = engine.counts()[Run.State.TERMINATED]
        while done < num_jobs and current_iter < max_iter:
            engine.progress()
            done = engine.counts()[Run.State.TERMINATED]
            current_iter += 1

        # check that they have been forgotten
        assert 0 == len(engine._managed.done)
        for task in tasks:
            assert task.execution.state == 'TERMINATED'
            assert not task._attached


def test_engine_progress_collection():
    with temporary_engine() as engine:
        seq = SimpleSequentialTaskCollection(3)
        engine.add(seq)

        # run through sequence
        while seq.execution.state != 'TERMINATED':
            engine.progress()
        assert seq.stage().jobname == 'stage2'
        assert seq.stage().execution.state == 'TERMINATED'


def test_engine_progress_collection_and_forget_terminated():
    with temporary_engine() as engine:
        engine.forget_terminated = True

        seq = SimpleSequentialTaskCollection(3)
        engine.add(seq)

        # run through sequence
        while seq.execution.state != 'TERMINATED':
            engine.progress()

        assert 0 == len(engine._managed.done)
        assert not seq._attached
        for task in seq.tasks:
            assert not task._attached


def test_engine_kill_SequentialTaskCollection():
    with temporary_engine() as engine:
        seq = SimpleSequentialTaskCollection(3)
        engine.add(seq)

        while seq.execution.state != 'RUNNING':
            engine.progress()

        # When the sequence is in RUNNING state, so must the first app
        assert seq.tasks[0].execution.state == 'RUNNING'
        assert seq.tasks[1].execution.state == 'NEW'
        assert seq.tasks[2].execution.state == 'NEW'

        # Killing a sequential should put all the applications in
        # TERMINATED state. However, we will need an extra run of
        # engine.progress() to update the status of all the jobs.
        engine.kill(seq)
        assert seq.tasks[0].execution.state == 'RUNNING'
        assert seq.tasks[1].execution.state == 'NEW'
        assert seq.tasks[2].execution.state == 'NEW'
        assert seq.execution.state == 'RUNNING'

        engine.progress()

        assert seq.tasks[0].execution.state == 'TERMINATED'
        assert seq.tasks[1].execution.state == 'TERMINATED'
        assert seq.tasks[2].execution.state == 'TERMINATED'
        assert seq.execution.state == 'TERMINATED'


def test_engine_kill_redo_SequentialTaskCollection():
    with temporary_engine() as engine:
        seq = SimpleSequentialTaskCollection(3)
        engine.add(seq)

        while seq.execution.state != 'RUNNING':
            engine.progress()

        # When the sequence is in RUNNING state, so must the first app
        assert seq.tasks[0].execution.state == 'RUNNING'
        assert seq.tasks[1].execution.state == 'NEW'
        assert seq.tasks[2].execution.state == 'NEW'

        # Killing a sequential should put all the applications in
        # TERMINATED state. However, we will need an extra run of
        # engine.progress() to update the status of all the jobs.
        engine.kill(seq)
        assert seq.tasks[0].execution.state == 'RUNNING'
        assert seq.tasks[1].execution.state == 'NEW'
        assert seq.tasks[2].execution.state == 'NEW'
        assert seq.execution.state == 'RUNNING'

        engine.progress()
        assert seq.tasks[0].execution.state == 'TERMINATED'
        assert seq.tasks[1].execution.state == 'TERMINATED'
        assert seq.tasks[2].execution.state == 'TERMINATED'
        assert seq.execution.state == 'TERMINATED'

        engine.redo(seq)
        assert seq.tasks[0].execution.state == 'NEW'
        assert seq.tasks[1].execution.state == 'NEW'
        assert seq.tasks[2].execution.state == 'NEW'
        assert seq.execution.state == 'NEW'

        engine.progress()
        assert seq.tasks[0].execution.state == 'SUBMITTED'
        assert seq.tasks[1].execution.state == 'NEW'
        assert seq.tasks[2].execution.state == 'NEW'


def test_engine_kill_ParallelTaskCollection():
    # Creates an engine with 2 cores.
    with temporary_engine(max_cores=2) as engine:
        par = SimpleParallelTaskCollection(3)
        engine.add(par)

        for _ in range(20):
            engine.progress()
            if par.execution.state == 'RUNNING':
                break

        # Because of our noop engine, as soon as the parallel is in
        # running we will have all jobs in SUBMITTED and the others in
        # NEW.
        assert (
            ['SUBMITTED', 'SUBMITTED', 'NEW'] ==
            [i.execution.state for i in par.tasks])

        # Killing a parallel should put all the applications in
        # TERMINATED state. However, we need two runs of
        # engine.progress() to update the status of all the jobs
        engine.kill(par)
        assert (
            ['SUBMITTED', 'SUBMITTED', 'NEW'] ==
            [i.execution.state for i in par.tasks])

        engine.progress()
        engine.progress()

        assert (
            ['TERMINATED', 'TERMINATED', 'TERMINATED'] ==
            [i.execution.state for i in par.tasks])
        assert par.execution.state == 'TERMINATED'


def test_engine_redo_SequentialTaskCollection():
    with temporary_engine() as engine:
        seq = SimpleSequentialTaskCollection(3)
        engine.add(seq)

        # run through sequence
        while seq.execution.state != 'TERMINATED':
            engine.progress()
        assert seq.stage().jobname == 'stage2'
        assert seq.stage().execution.state == 'TERMINATED'

        engine.redo(seq, 1)
        assert seq.stage().jobname == 'stage1'
        assert seq.stage().execution.state == 'NEW'

        # run through sequence again
        while seq.execution.state != 'TERMINATED':
            engine.progress()
        assert seq.stage().jobname == 'stage2'
        assert seq.stage().execution.state == 'TERMINATED'


def test_engine_redo_ParallelTaskCollection():
    with temporary_engine() as engine:
        par = SimpleParallelTaskCollection(5)
        engine.add(par)

        # run until terminated
        while par.execution.state != Run.State.TERMINATED:
            engine.progress()

        engine.redo(par)
        assert par.execution.state == Run.State.NEW
        for task in par.tasks:
            assert task.execution.state == Run.State.NEW


def test_engine_redo_Task1():
    """Test correct use of `Engine.redo()` with a `Task` instance."""
    with temporary_engine() as engine:
        task = SuccessfulApp()
        engine.add(task)

        # run until terminated
        while task.execution.state != Run.State.TERMINATED:
            engine.progress()
        assert task.execution.state == Run.State.TERMINATED

        # no do it all over again
        engine.redo(task)
        assert task.execution.state == Run.State.NEW

        engine.progress()
        assert task.execution.state != Run.State.NEW
        assert task.execution.state in [Run.State.SUBMITTED, Run.State.RUNNING]


def test_engine_redo_Task2():
    """Test that `Engine.redo()` raises if called on a Task that is not TERMINATED."""
    with temporary_engine() as engine:
        task = SuccessfulApp()
        engine.add(task)

        engine.progress()
        assert task.execution.state != Run.State.NEW

        # cannot redo a task that is not yet terminated
        with pytest.raises(AssertionError):
            task.redo()
            pytest.fail("`Task.redo()` succeeded on task not yet finished")


def test_engine_redo_Task3():
    """Test that `Engine.redo()` is a no-op if the Task is still NEW."""
    with temporary_engine() as engine:
        task = SuccessfulApp()
        engine.add(task)
        task.redo()


def test_engine_resubmit():
    with temporary_engine() as engine:
        app = SuccessfulApp()
        engine.add(app)

        # run through sequence
        while app.execution.state != 'TERMINATED':
            engine.progress()

        engine.submit(app, resubmit=True)
        assert app.execution.state == 'NEW'

        # run through sequence again
        while app.execution.state != 'TERMINATED':
            engine.progress()
        assert app.execution.state == 'TERMINATED'


def test_engine_submit1():
    """Engine.submit is equivalent to `add` if a task is not yet managed."""
    with temporary_engine() as engine:
        assert engine.counts()['NEW'] == 0

        app = SuccessfulApp()
        assert app.execution.state == 'NEW'
        engine.submit(app)
        assert app.execution.state == 'NEW'
        assert engine.counts()['NEW'] == 1

        engine.progress()
        assert app.execution.state in ['SUBMITTED', 'RUNNING']
        assert engine.counts()['NEW'] == 0
        assert engine.counts()[app.execution.state] == 1


def test_engine_submit2():
    """Engine.submit is a no-op if a task is already managed."""
    with temporary_engine() as engine:
        app = SuccessfulApp()
        engine.add(app)
        assert engine.counts()['NEW'] == 1

        engine.submit(app)
        assert engine.counts()['NEW'] == 1

        engine.progress()
        state = app.execution.state
        assert state in ['SUBMITTED', 'RUNNING']
        assert engine.counts()['NEW'] == 0
        assert engine.counts()[state] == 1

        engine.submit(app)
        assert app.execution.state == state
        assert engine.counts()[state] == 1


def test_engine_submit_to_multiple_resources(num_resources=3, num_jobs=50):
    """Test job spread across multiple resources."""
    # sanity check for parameters
    assert num_jobs > 10*(num_resources*(num_resources-1)/2), \
        "There must be enough jobs to fill the first N-1 resources"
    assert num_jobs < 10*(num_resources*(num_resources+1)/2), \
        "Too many jobs: would fill all resources"
    # set up
    cfg = gc3libs.config.Configuration()
    cfg.TYPE_CONSTRUCTOR_MAP['noop'] = ('gc3libs.backends.noop', 'NoOpLrms')
    for n in range(num_resources):
        name = 'test{nr}'.format(nr=n+1)
        cfg.resources[name].update(
            name=name,
            type='noop',
            auth='none',
            transport='local',
            max_cores_per_job=1,
            max_memory_per_core=1*GB,
            max_walltime=8*hours,
            max_cores=((n+1)*10),
            architecture=Run.Arch.X86_64,
        )
    core = Core(cfg)
    engine = Engine(core)
    # generate 50 no-op tasks
    for n in range(num_jobs):
        name = 'app{nr}'.format(nr=n)
        engine.add(SuccessfulApp(name))
    # submit them all
    engine.progress()
    # get handles to the actual backend objects
    rscs = [
        core.get_backend('test{nr}'.format(nr=n+1))
        for n in range(num_resources)
    ]
    num_jobs_per_resource = [
        len([task for task in engine._managed.to_update
             if task.execution.resource_name == rsc.name])
        for rsc in rscs
    ]
    # check that all jobs have been submitted and that each
    # resource got at least one job
    assert sum(num_jobs_per_resource) == num_jobs
    for num in num_jobs_per_resource:
        assert num > 0
    # since TYPE_CONSTRUCTOR_MAP is a class-level variable, we
    # need to clean up otherwise other tests will see the No-Op
    # backend
    del cfg.TYPE_CONSTRUCTOR_MAP['noop']


def test_create_engine_default():
    """Test `create_engine` with factory defaults."""
    with temporary_config_file() as cfgfile:
        # std factory params
        engine = create_engine(cfgfile.name)
        assert isinstance(engine, Engine)


def test_create_engine_non_default1():
    """Test `create_engine` with one non-default argument."""
    with temporary_config_file() as cfgfile:
        engine = create_engine(cfgfile.name, can_submit=False)
        assert engine.can_submit == False


def test_create_engine_non_default2():
    """Test `create_engine` with several non-default arguments."""
    with temporary_config_file() as cfgfile:
        engine = create_engine(cfgfile.name,
                               can_submit=False,
                               max_in_flight=1234)
        assert engine.can_submit == False
        assert engine.max_in_flight == 1234


def test_create_engine_with_core_options():
    """Test `create_engine` with a mix of Engine and Core options."""
    with temporary_config_file() as cfgfile:
        # use a specific MatchMaker instance for equality testing
        mm = MatchMaker()
        engine = create_engine(cfgfile.name,
                               can_submit=False,
                               matchmaker=mm,
                               auto_enable_auth=False)
        assert engine.can_submit == False
        assert engine._core.matchmaker == mm
        assert engine._core.auto_enable_auth == False


def test_engine_find_task_by_id():
    """
    Test that saved tasks are can be retrieved from the Engine given their ID only.
    """
    with temporary_core() as core:
        with temporary_directory() as tmpdir:
            store = FilesystemStore(tmpdir)
            engine = Engine(core, store=store)

            task = SuccessfulApp()
            store.save(task)
            engine.add(task)

            task_id = task.persistent_id
            assert engine.find_task_by_id(task_id) == task


def test_engine_cannot_find_task_by_id_if_not_saved():
    """
    Test that *unsaved* tasks are cannot be retrieved from the Engine given their ID only.
    """
    with temporary_core() as core:
        with temporary_directory() as tmpdir:
            store = FilesystemStore(tmpdir)
            engine = Engine(core, store=store)

            task = SuccessfulApp()
            engine.add(task)

            store.save(task)  # guarantee it has a `.persistent_id`
            task_id = task.persistent_id
            with pytest.raises(KeyError):
                engine.find_task_by_id(task_id)


def test_engine_cannot_find_task_by_id_if_no_store():
    """
    Test that `Engine.find_task_by_id` always raises `KeyError` if the Engine has no associated store.
    """
    with temporary_engine() as engine:
       with temporary_directory() as tmpdir:
            store = FilesystemStore(tmpdir)

            task = SuccessfulApp()
            engine.add(task)

            store.save(task)  # guarantee it has a `.persistent_id`
            task_id = task.persistent_id
            with pytest.raises(KeyError):
                engine.find_task_by_id(task_id)


@pytest.mark.parametrize("limit_submitted,limit_in_flight", [
    (2, 10),
    (10, 5),
])
def test_engine_limits(limit_submitted, limit_in_flight,
                       num_jobs=30, max_iter=100):
    """
    Test that `Engine.limit_in_flight` and `Engine.limit_submitted` are honored.
    """
    with temporary_engine(max_cores=50) as engine:
        # set limits
        engine.max_in_flight = 10
        engine.max_submitted = 2
        # populate with test apps
        apps = []
        for n in range(num_jobs):
            name = 'app{nr}'.format(nr=n)
            app = SuccessfulApp(name)
            engine.add(app)
            apps.append(app)
        # run all apps for (up to) a fixed nr of steps
        iter_nr = 0
        stats = engine.counts()
        while stats['TERMINATED'] < num_jobs and iter_nr < max_iter:
            iter_nr += 1
            engine.progress()
            stats = engine.counts()
            submitted = stats['SUBMITTED']
            assert submitted <= engine.max_submitted
            in_flight = (stats['SUBMITTED'] + stats['RUNNING'])
            assert in_flight <= engine.max_in_flight
        # catch errors in case of termination because of exceeded iter count
        assert stats["TERMINATED"] == num_jobs


@pytest.mark.parametrize("max_cores", [1, 2, 5, 12, 24])
def test_engine_counts1(max_cores, num_jobs=24, max_iter=1000):
    """
    Test that `Engine.count()` returns correct results with a plain list of tasks.
    """
    def populate(engine):
        apps = []
        for n in range(num_jobs):
            name = 'app{nr}'.format(nr=n)
            app = SuccessfulApp(name)
            apps.append(app)
            engine.add(app)
        return apps
    _test_engine_counts(populate, max_cores, num_jobs, max_iter=max_iter)

@pytest.mark.parametrize("max_cores", [1, 2, 5, 12, 24])
def test_engine_counts2(max_cores, n1=6, n2=6, max_iter=1000):
    """
    Test that `Engine.count()` returns correct results with a list of parallel collections.
    """
    def populate(engine):
        apps = []
        for n in range(n1):
            par = SimpleParallelTaskCollection(n2)
            par.jobname = 'par{nr}'.format(nr=n)
            engine.add(par)
            apps.append(par)
            for task in par.tasks:
                apps.append(task)
        return apps
    _test_engine_counts(populate, max_cores, n1 + n1*n2, max_iter=max_iter)

def _test_engine_counts(populate, max_cores, num_jobs,
                        num_new_jobs=None, max_iter=100):
    """
    Common code for the `test_engine_counts*` tests.
    """
    # make the `.progress()` call a little less deterministic
    transition_graph = {
        Run.State.SUBMITTED:   {0.8: Run.State.RUNNING},
        Run.State.RUNNING:     {0.8: Run.State.TERMINATING},
        Run.State.TERMINATING: {0.8: Run.State.TERMINATED},
    }
    with temporary_engine(transition_graph, max_cores=max_cores) as engine:
        # populate with test apps
        apps = populate(engine)

        # initial check on stats
        actual_counts = engine.counts()
        if num_new_jobs is None:
            num_new_jobs = num_jobs
        assert actual_counts['NEW'] == num_new_jobs

        # check
        iter_nr = 0
        while (actual_counts['TERMINATED'] < num_jobs
               and iter_nr < max_iter):
            iter_nr += 1
            engine.progress()
            actual_counts = engine.counts()
            expected_counts = _compute_counts(apps)
            _check_counts(actual_counts, expected_counts)

def _compute_counts(apps):
    """
    Helper method for `test_*_counts`.

    Explicitly compute the job/state counts by running through the
    entire list of jobs.
    """
    result = defaultdict(int)
    for app in apps:
        result['total'] += 1
        state = app.execution.state
        result[state] += 1
        if state == 'TERMINATED':
            if app.execution.returncode == 0:
                result['ok'] += 1
            else:
                result['failed'] += 1
    return result

def _check_counts(actual, expected):
    """
    Common code for `test_*_counts`.

    Check that `actual` is internally consistent and that it agrees
    with `expected` on every task state.
    """
    total = expected['total']
    # internal consistency checks
    assert actual['total'] == total
    assert total == sum([
        actual['NEW'],
        actual['SUBMITTED'],
        actual['RUNNING'],
        actual['TERMINATING'],
        actual['TERMINATED'],
    ])
    assert actual['TERMINATED'] == (
        actual['ok'] + actual['failed'])
    # compare with explicit counting
    for state in [
            'NEW', 'SUBMITTED', 'RUNNING', 'TERMINATING',
            'TERMINATED', 'ok', 'failed',
    ]:
        # need to make our own error message, otherwise
        # `pytest` just prints something like `assert 0 == 2`
        try:
            assert actual[state] == expected[state]
        except AssertionError:
            raise AssertionError(
                "Actual count for `{0}` is {1} but expected {2}"
                .format(state, actual[state], expected[state]))


if __name__ == "__main__":
    import pytest
    pytest.main(["-v", __file__])
